from contextlib import contextmanager
from typing import Tuple, Union
from time import perf_counter

import numpy as np
import pylinalg as la

from ..cameras import Camera, PerspectiveCamera
from ..cameras._perspective import fov_limit, fov_distance_factor
from ..renderers import Renderer
from ..utils.viewport import Viewport


class Controller:
    """The base camera controller.

    The purpose of a controller is to provide an API to control a camera,
    and to convert user (mouse) events into camera adjustments.

    Parameters
    ----------
    camera: Camera
        The camera to control (optional). Must be a perspective or orthographic camera.
    enabled: bool
        Whether the controller is enabled (i.e. responds to events).
    damping: float
        The amount of motion damping. Zero is no damping, 10 a lot. Default 4.
    auto_update: bool
        When True (default), the controller is pretty much plug-and-play.
        For more control, it can be set to False. The controller will
        then not update the cameras automatically, and will not request
        new draws. You will then need to periodically call ``tick()``, and
        use the return value (a camera state dict).
    register_events: Renderer or Viewport
        If given and not None, will call ``.register_events()``.

    Usage
    -----

    There are multiple ways that the controller can be used.
    The easiest (and most common) approach is to use the Pygfx event system and
    make the controller listen to viewport events (using ``register_events``).

    An alternative is to feed your own events into the ``handle_event()``
    method. You'd have to mimic or use Pygfx event objects.

    The controller can also be used programmatically by calling "action methods"
    like ``pan()``, ``zoom()`` and ``rotate()``.
    """

    _default_controls = {}

    def __init__(
        self,
        camera=None,
        *,
        enabled=True,
        damping=4,
        auto_update=True,
        register_events=None,
    ):
        # Init cameras
        self._cameras = []  # tuples of camera + filters
        if camera is not None:
            self.add_camera(camera)

        self._linked_controllers = []

        # Props
        self.enabled = enabled
        self.damping = damping
        self.auto_update = auto_update

        # Init controls config
        action_names = set(
            action_tuple[0] for action_tuple in self._default_controls.values()
        )
        self._controls = Controls(*action_names)
        self._controls.update(self._default_controls)

        # State info used during interactions.
        self._actions = {}
        self._last_cam_state = {}
        self._last_tick_time = 0

        # Maybe register events. The register_events arg must be a renderer or viewport.
        self._call_later = None
        if register_events is not None:
            self.register_events(register_events)

    @property
    def cameras(self):
        """A tuple with the cameras under control, in the order that they were added."""
        return tuple(cam for cam, include, exclude in self._cameras)

    def add_camera(self, camera, *, include_state=None, exclude_state=None):
        """Add a camera to control.

        If the camera is already registered, it only updates the include_state
        and exclude_state (without changing the order of the cameras).

        The ``include_state`` and ``exclude_state`` arguments represent the
        camera state fields to include/exclude, when updating this camera. This
        can be used to "partially link" a camera. These args are None by default
        (i.e. no filtering).

        See ``camera.get_state()`` for available fields. Useful state names for
        the perspective and orthographhic camera are: 'x', 'y', 'z' (or
        'position' for all three), 'width', 'height', 'rotation', 'zoom', and
        'depth_range'.
        """
        if not isinstance(camera, Camera):
            raise TypeError("Controller.add_camera expects a Camera object.")
        if not isinstance(camera, PerspectiveCamera):
            raise TypeError(
                "Controller.add_camera expects a perspective or orthographic camera."
            )

        if include_state is not None:
            if not isinstance(include_state, set):
                raise TypeError("add_camera() include_state must be a set.")
            include_state = set(include_state)  # copy
            if "position" in include_state:
                include_state.discard("position")
                include_state.update({"x", "y", "z"})

        if exclude_state is not None:
            if not isinstance(exclude_state, set):
                raise TypeError("add_camera() exclude_state must be a set.")
            exclude_state = set(exclude_state)  # copy
            if "position" in exclude_state:
                exclude_state.discard("position")
                exclude_state.update({"x", "y", "z"})

        new_entry = camera, include_state, exclude_state

        # Replace or append
        for i in range(len(self._cameras)):
            if self._cameras[i][0] is camera:
                self._cameras[i] = new_entry
                break
        else:
            self._cameras.append(new_entry)

    def remove_camera(self, camera):
        """Remove a camera from the list of cameras to control.

        If the camera is not registered, this does nothing.
        """
        if not isinstance(camera, Camera):
            raise TypeError("Controller.remove_camera expects a Camera object.")
        new_cameras = []
        for cam, include_state, exclude_state in self._cameras:
            if cam is not camera:
                new_cameras.append((cam, include_state, exclude_state))
        self._cameras = new_cameras
        if not self._cameras:
            self._actions = {}  # cancel any in-progress actions

    @property
    def enabled(self):
        """Whether the controller responds to events."""
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        self._enabled = bool(value)
        if not self._enabled:
            self._actions = {}  # cancel any in-progress actions

    @contextmanager
    def pause(self):
        """
        Context manager to temporarily disable the controller. Controller is set back to the original enabled/disabled
        state when leaving the context.

        Usage
        -----

        with controller.pause():
            # do things while controller.enabled is False

        # outside context manager, controller.enabled is set back to the original value

        """
        is_enabled = self.enabled

        self.enabled = False

        try:
            yield
        finally:
            self.enabled = is_enabled

    @property
    def damping(self):
        """The amount of motion damping (i.e. smoothing).
        Lower values dampen less. Typical values would be below 10.
        """
        return self._damping

    @damping.setter
    def damping(self, value):
        self._damping = max(0.0, float(value))

    @property
    def auto_update(self):
        """Whether the controller automatically requests a new draw at the canvas."""
        return self._auto_update

    @auto_update.setter
    def auto_update(self, value):
        self._auto_update = bool(value)

    @property
    def controls(self):
        """A dictionary that maps buttons/keys to actions.

        Can be modified to configure how the controller reacts to
        events. Possible keys include 'mouse1' to 'mouse5', single
        characters for the corresponding key presses, and 'arrowleft',
        'arrowright', 'arrowup', 'arrowdown', 'tab', 'enter', 'escape',
        'backspace', and 'delete'.

        Each action value is a tuple with 3 fields:

        * The `action name`, e.g. 'pan', see ``controller.controls.action_names``
          for the possible names.
        * The `mode`: 'drag', 'push', 'peek', 'repeat'. Drag represents mouse drag,
          push means the action is performed as the key is pushed, peek
          means that the action is undone once the key is released, and repeat
          means that while the key is held down, the value updates with the given
          amount per second.
        * The `multiplier` is the value that the set value from the event is
          multiplied with.

        """
        return self._controls

    @controls.setter
    def controls(self, value):
        self._controls.clear()
        self._controls.update(value)

    def register_events(self, viewport_or_renderer: Union[Viewport, Renderer]):
        """Apply the default interaction mechanism to a render canvas.
        Needs either a viewport or renderer.
        """
        viewport = Viewport.from_viewport_or_renderer(viewport_or_renderer)
        viewport.renderer.add_event_handler(
            lambda event: self.handle_event(event, viewport),
            "pointer_down",
            "pointer_move",
            "pointer_up",
            "key_down",
            "key_up",
            "wheel",
            "before_render",
        )

    def tick(self):
        """Should be called periodically to keep the camera up to date.

        If ``auto_update`` is True, this is done automatically. Returns
        a dict with the new camera state, or None if there are no
        running actions.
        """

        if not self._actions:
            return None

        # Get elapsed time since last frame
        now = perf_counter()
        elapsed_time = now - self._last_tick_time
        self._last_tick_time = now
        # Determine damping/smoothing factor to update action values.
        # In the formula below the mul with elapsed_time is equivalent
        # to a division by fps. The mul with 50 is just so typical
        # values for damping lay between 0..10 :)
        factor = 1
        if self.damping > 0:
            factor = min(1, 50 * elapsed_time / self.damping)

        to_pop = []
        for key, action in self._actions.items():
            if action.mode == "repeat":
                action.increase_target(elapsed_time)
            action.tick(factor)
            if action.is_at_target and action.done:
                to_pop.append(key)
            self._apply_action(action)

        # Remove actions that are done
        for key in to_pop:
            self._actions.pop(key)

        return self._update_cameras()

    def _get_target_vec(self, camera_state, **kwargs):
        """Method used by the controller implementations to determine the "target"."""
        rotation = kwargs.get("rotation", camera_state["rotation"])
        extent = 0.5 * (camera_state["width"] + camera_state["height"])
        extent = kwargs.get("extent", extent)
        fov = kwargs.get("fov", camera_state.get("fov"))

        distance = fov_distance_factor(fov_limit(fov)) * extent
        return la.vec_transform_quat((0, 0, -distance), rotation)

    def _get_camera_vecs(self, rect):
        """Get vectors orthogonal to camera axii."""
        if not self._cameras:
            raise ValueError("No cameras attached!")

        camera, _, _ = self._cameras[0]
        cam_state = self._get_camera_state()
        target = cam_state["position"] + self._get_target_vec(cam_state)
        vecx, vecy = get_screen_vectors_in_world_cords(target, rect[2:], camera)

        return vecx, vecy

    def _get_camera_state(self):
        """Gets the first camera's state, or the internal state if
        the controller is currently interacting.
        """
        if self._actions:
            return self._last_cam_state.copy()
        elif self._cameras:
            return self._cameras[0][0].get_state()
        else:
            raise ValueError("No cameras attached!")

    def _set_camera_state(self, new_state):
        """Set the internal camera state. Camera state can be updated multiple times
        before updating the cameras.
        """
        # Update the state dict, turn arrays into tuples, because
        # downstream code might expect that.
        for k, v in new_state.items():
            if isinstance(v, np.ndarray):
                v = tuple(v)
            self._last_cam_state[k] = v

    def _update_cameras(self):
        """Update the cameras using the internally stored state. Should only
        be called by code that knows that internally stored state is valid.
        """
        if self._auto_update:
            for camera, include_state, exclude_state in self._cameras:
                state = self._last_cam_state
                if include_state or exclude_state:
                    if "position" in state:
                        state = state.copy()
                        state["x"], state["y"], state["z"] = state.pop("position")
                    if include_state is not None:
                        state = {k: v for k, v in state.items() if k in include_state}
                    if exclude_state is not None:
                        state = {
                            k: v for k, v in state.items() if k not in exclude_state
                        }
                camera.set_state(state)
        return self._last_cam_state

    def add_default_event_handlers(self, *args):
        raise DeprecationWarning(
            "controller.add_default_event_handlers(viewport, camera) -> controller.register_events(viewport)"
        )

    def update_camera(self, *args):
        raise DeprecationWarning("controller.update_camera() is no longer necessary")

    # %% Builtin event handling

    def handle_event(self, event, viewport):
        if not self.enabled:
            return
        if not self._cameras:
            return

        rect = viewport.rect
        need_update = False

        type = event.type
        if type.startswith(("pointer_", "key_", "wheel")):
            modifiers = {m.lower() for m in event.modifiers}
            if type.startswith("key_"):
                modifiers.discard(event.key.lower())
            modifiers_prefix = "+".join([*sorted(modifiers), ""])

        if type == "before_render":
            # Do a tick, updating all actions, and using them to update the camera state.
            # Note that tick() removes actions that are done and have reached the target.
            if self._auto_update and self._actions:
                self.tick()
                need_update = True
        elif type == "pointer_down" and viewport.is_inside(event.x, event.y):
            # Start a drag, or an action with mode push/peek/repeat
            key = modifiers_prefix + f"mouse{event.button}"
            action_tuple = self._controls.get(key)
            if action_tuple:
                need_update = True
                if action_tuple[1] == "drag":
                    # Dont start a new drag if there is one going
                    if not any(
                        (a.mode == "drag" and not a.done)
                        for a in self._actions.values()
                    ):
                        pos = event.x, event.y
                        self._create_action(key, action_tuple, pos, pos, rect)
                else:
                    self._handle_button_down(key, action_tuple, viewport)
        elif type == "pointer_move":
            # Update all drag actions
            for action in self._actions.values():
                if action.mode == "drag" and not action.done:
                    action.set_target(np.array((event.x, event.y)))
                    need_update = True
        elif type == "pointer_up":
            # Stop all drag actions
            for action in self._actions.values():
                if action.mode == "drag":
                    action.done = True
            # End button presses, regardless of modifier state
            need_update = self._handle_button_up(f"mouse{event.button}")
        elif type == "wheel" and viewport.is_inside(event.x, event.y):
            # Wheel events. Technically there is horizontal and vertical scroll,
            # but this does not work well cross-platform, so we consider it 1D.
            key = modifiers_prefix + "wheel"
            action_tuple = self._controls.get(key)
            if action_tuple:
                need_update = True
                d = event.dy or event.dx
                pos = event.x, event.y
                action = self._actions.get(key, None)
                if action is None:
                    action = self._create_action(key, action_tuple, 0, pos, rect)
                    action.done = True
                action.increase_target(d)
        elif type == "key_down":
            # Start an action with mode push/peek/repeat
            key = modifiers_prefix + f"{event.key.lower()}"
            action_tuple = self._controls.get(key)
            if action_tuple:
                need_update = True
                self._handle_button_down(key, action_tuple, viewport)
        elif type == "key_up":
            # End key presses, regardless of modifier state
            need_update = self._handle_button_up(f"{event.key.lower()}")

        if need_update and self.auto_update:
            viewport.renderer.request_draw()

    def _handle_button_down(self, key, action_tuple, viewport):
        """Common code to handle key/mouse button presses."""
        mode = action_tuple[1]
        action = self._actions.get(key, None)
        if action is None:
            action = self._create_action(key, action_tuple, 0, None, viewport.rect)
            action.snap_distance = 0.01
        action.done = mode == "push"
        if mode in ("push", "peek"):
            action.set_target(1)

    def _handle_button_up(self, button):
        """Common code to handle key/mouse button releases."""
        need_update = False
        for key, action in self._actions.items():
            if key == button or key.endswith("+" + button):
                need_update = True
                action.done = True
                if action.mode == "peek":
                    action.set_target(action.target_value * 0)
        return need_update

    def _create_action(self, key, action_tuple, offset, screen_pos, rect):
        """Creates an action object, which helps keep track of the operation."""

        key = key or str(perf_counter())
        if screen_pos is None:
            screen_pos = rect[0] + rect[2] / 2, rect[1] + rect[3] / 2

        # Get vectors orthogonal to camera axii, scaled by pixel unit
        vecx, vecy = self._get_camera_vecs(rect)

        # Make sure that we have an up-to-date cam_state
        if not self._actions:
            self._last_cam_state = self._cameras[0][0].get_state()
            self._last_tick_time = perf_counter()

        # Create action
        kwargs = dict(rect=rect, screen_pos=screen_pos, vecx=vecx, vecy=vecy)
        action = Action(action_tuple, offset, kwargs)

        self._actions[key] = action
        return action

    def _apply_action(self, action):
        """Apply the action by calling the appropriate update method."""
        # Get function to call
        func = getattr(self, "_update_" + action.name)

        # Collect the kwargs that the function needs
        code = func.__func__.__code__
        assert code.co_argcount == 2  # self and delta
        assert code.co_posonlyargcount == 0
        kwargnames = code.co_varnames[2 : 2 + code.co_kwonlyargcount]
        kwargs = {k: action.kwargs[k] for k in kwargnames}

        # Call it!
        func(action.delta, **kwargs)

    # %% Actions on the base class

    def quickzoom(self, delta: float, *, animate=False):
        """Zoom the view using the camera's zoom property. This is intended
        for temporary zoom operations.

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("quickzoom", "push", 1.0)
            action = self._create_action(None, action_tuple, 0.0, None, (0, 0, 1, 1))
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            self._update_quickzoom(delta)
            return self._update_cameras()

    def _update_quickzoom(self, delta):
        assert isinstance(delta, (int, float))
        zoom = self._get_camera_state()["zoom"]
        new_cam_state = {"zoom": zoom * 2**delta}
        self._set_camera_state(new_cam_state)

    def update_fov(self, delta, *, animate):
        """Adjust the field of view with the given delta value (Limited to [1, 179]).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("fov", "push", 1.0)
            action = self._create_action(None, action_tuple, 0.0, None, (0, 0, 1, 1))
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            self._update_fov(delta)
            return self._update_cameras()

    def _update_fov(self, delta: float):
        fov_range = self._cameras[0][0]._fov_range

        # Get current state
        cam_state = self._get_camera_state()
        position = cam_state["position"]
        fov = cam_state["fov"]

        # Update fov and position
        new_fov = min(max(fov + delta, fov_range[0]), fov_range[1])
        new_fov = fov_limit(new_fov)
        pos2target1 = self._get_target_vec(cam_state, fov=fov)
        pos2target2 = self._get_target_vec(cam_state, fov=new_fov)
        new_position = position + pos2target1 - pos2target2

        self._set_camera_state({"fov": new_fov, "position": new_position})


class Action:
    """Simple value to represent a value to change an action with."""

    def __init__(self, action_tuple, offset=0.0, kwargs=None):
        action_name, mode, multiplier = action_tuple

        # The name of the action, used to dispatch to controller._update_xx()
        self.name = action_name

        # The offset defines the dimension of the input
        self.offset = self._clean_up_value(offset)

        # The multiplier defines the dimension of the output.
        # If it's less than the input, the first input dimension(s) are dropped.
        self.multiplier = self._clean_up_value(multiplier)

        # Derive the zero value
        if isinstance(offset, float):
            zero = 0.0
        else:
            zero = np.zeros_like(offset)

        # Init the values
        self.last_value = zero
        self.target_value = zero
        self.current_value = zero
        self.repeat_multiplier = 0

        self.snap_distance = 0.5
        self.done = False
        self.mode = mode
        self.kwargs = kwargs or {}

    def _clean_up_value(self, value):
        # Turns value into either a float or a flat nd float array
        if isinstance(value, (int, float)):
            return float(value)
        elif isinstance(value, (list, tuple)):
            return np.array(value, dtype=np.float64)
        elif isinstance(value, np.ndarray):
            if value.size == 1:
                return float(value)
            else:
                return value.flatten()  # makes copy

    def __repr__(self):
        return f"<Action '{self.name}' {self.current_value}>"

    def set_target(self, value):
        self.target_value = value - self.offset

    def increase_target(self, value):
        if self.mode == "repeat":
            # We increase/decrease a multiplier linearly
            lag_time = 1.5  # seconds
            if not self.done:
                self.repeat_multiplier = min(
                    1.0, self.repeat_multiplier + value / lag_time
                )
            else:
                self.repeat_multiplier = max(
                    0.0, self.repeat_multiplier - value / lag_time
                )
            self.target_value = self.target_value + value * self.repeat_multiplier
        else:
            self.target_value = self.target_value + value

    def tick(self, factor=1):
        # Update value
        new_value = (1 - factor) * self.current_value + factor * self.target_value
        self.last_value = self.current_value
        self.current_value = new_value
        #
        dist_to_target = np.abs(self.target_value - self.current_value).max()
        if dist_to_target < self.snap_distance:
            self.current_value = self.target_value * 1.0  # make a copy if array

    @property
    def delta(self):
        """Get the delta value, multiplied with the multiplier."""
        delta = self.multiplier * (self.current_value - self.last_value)
        if not isinstance(delta, float):
            delta = tuple(delta)
            if isinstance(self.multiplier, float):
                delta = delta[-1]
            elif self.multiplier.size < len(delta):
                delta = delta[-self.multiplier.size :]
        return delta

    @property
    def is_at_target(self):
        if self.mode == "repeat":
            if self.repeat_multiplier:
                return False
        return np.all(self.current_value == self.target_value)


class Controls(dict):
    """Overloaded dict so we can validate when an item is set."""

    _buttons = "mouse1", "mouse2", "mouse3", "mouse4", "mouse5", "wheel"
    _buttons += "arrowleft", "arrowright", "arrowup", "arrowdown"
    _buttons += "tab", "enter", "escape", "backspace", "delete"
    _buttons += "shift", "control"

    _modes = "drag", "push", "peek", "repeat"

    def __init__(self, *actions):
        self._actions = tuple(actions)

    def __repr__(self):
        # Pretty print using one rule per line
        if not self:
            return "{}"
        s = "{"
        for key, action_tuple in self.items():
            s += f"\n    '{key}': {action_tuple!r},"
        s += "\n}"
        return s

    @property
    def action_names(self):
        """The possible action names."""
        return self._actions

    def __setitem__(self, key, action_tuple):
        # Check the button
        if not isinstance(key, str):
            raise TypeError("Controls key must be str")
        *modifiers, button = key.split("+")
        modifiers = sorted([m.lower() for m in modifiers])
        button = button.lower()
        for m in modifiers:
            if m not in ("shift", "control", "alt"):
                raise ValueError(f"Invalid key modifier '{m}'")
        if len(button) == 1:
            pass  # a key
        elif button not in self._buttons:
            raise ValueError(
                f"Invalid button/key '{button}', pick a char, or one of {self._buttons}"
            )
        # Check the action
        if not (isinstance(action_tuple, (list, tuple)) and len(action_tuple) == 3):
            raise TypeError("Controls action must be 3-element tuples")
        action, mode, multiplier = action_tuple
        if action not in self._actions:
            raise ValueError(f"Invalid action '{action}', pick one of {self._actions}")
        if mode not in self._modes:
            raise ValueError(f"Invalid mode '{mode}', pick one of {self._modes}")
        if mode == "drag" and not button.startswith("mouse"):
            raise ValueError("Drag mode only allowed for mouse buttons.")
        if button == "wheel" and mode != "push":
            raise ValueError("Only mode 'push' allowed with 'wheel'.")
        if isinstance(multiplier, (int, float)):
            multiplier = float(multiplier)
        elif isinstance(multiplier, (list, tuple)):
            multiplier = tuple(float(x) for x in multiplier)
        elif isinstance(multiplier, np.ndarray):
            if multiplier.size == 1:
                multiplier = float(multiplier)
            else:
                multiplier = tuple(float(x) for x in multiplier)
        # Store
        modifiers_prefix = "+".join([*modifiers, ""])
        key = modifiers_prefix + button
        super().__setitem__(key, (action, mode, multiplier))

    def setdefault(self, key, default):
        # Overloaded to make use of our implementation of __setitem__
        if key not in self:
            self[key] = default
        return self[key]

    def update(self, e, **f):
        # Overloaded to make use of our implementation of __setitem__
        for k, v in e.items():
            self[k] = v
        for k, v in f.items():
            self[k] = v


def get_screen_vectors_in_world_cords(
    center_world: Tuple[float, float, float],
    scene_size: Tuple[float, float],
    camera: Camera,
):
    """Given a reference center location (in 3D world coordinates)
    Get the vectors corresponding to the x and y direction in screen coordinates.
    These vectors are scaled so that they can simply be multiplied with the
    delta x and delta y.
    """

    # Linalg conv
    camera_world = camera.world.matrix
    camera_world_inverse = camera.world.inverse_matrix
    camera_projection = camera.projection_matrix
    camera_projection_inverse = camera.projection_matrix_inverse

    # Get center location on screen
    center_ndc = la.vec_transform(
        la.vec_transform(center_world, camera_world_inverse), camera_projection
    )

    # Step 1 NDC unit in x and y, and convert these positions back to world
    posx_ndc = center_ndc + np.array((1, 0, 0))
    posy_ndc = center_ndc + np.array((0, 1, 0))
    posx_world = la.vec_transform(
        la.vec_transform(posx_ndc, camera_projection_inverse),
        camera_world,
    )
    posy_world = la.vec_transform(
        la.vec_transform(posy_ndc, camera_projection_inverse),
        camera_world,
    )

    # Calculate the vectors, and scale to logical pixels.
    vecx_world = posx_world - center_world
    vecy_world = posy_world - center_world
    return vecx_world * 2.0 / scene_size[0], vecy_world * 2.0 / scene_size[1]



"""
Functions to update resources.
"""

import wgpu

from ....resources import Texture, Buffer
from ....utils import logger

from .utils import to_texture_format, GfxSampler, GfxTextureView
from .mipmapsutil import get_mip_level_count, generate_texture_mipmaps
from .shared import get_shared


# Alternative texture formats that we support by padding channels as needed.
# Maps virtual_format -> (wgpu_format, pad_value, nbytes)
ALTTEXFORMAT = {
    "rgb8snorm": ("rgba8snorm", 127, 1),
    "rgb8unorm": ("rgba8unorm", 255, 1),
    "rgb8sint": ("rgba8sint", 127, 1),
    "rgb8uint": ("rgba8uint", 255, 1),
    "rgb16sint": ("rgba16sint", 32767, 2),
    "rgb16uint": ("rgba16uint", 65535, 2),
    "rgb32sint": ("rgba32sint", 2147483647, 4),
    "rgb32uint": ("rgba32uint", 4294967295, 4),
    "rgb16float": ("rgba16float", 1, 2),
    "rgb32float": ("rgba32float", 1, 4),
}


def update_resource(resource):
    """Update the contents of a buffer or texture."""
    if isinstance(resource, Buffer):
        return _update_buffer(resource)
    elif isinstance(resource, Texture):
        return _update_texture(resource)
    else:
        raise ValueError(f"Invalid resource type: {resource.__class__.__name__}")


# Note on how buffer and texture updates work:
#
# * When a resource is first created, it's _wgpu_object attribute is None
#   and its _wgpu_flags is unset.
# * When the resource is actually being used somewhere, it will end up
#   in the logic that creates a pipeline object (e.g. in _pipeline.py), which
#   sets the appropriate usage flags (because that code knows how the resource
#   is used) and then uses ensure_wgpu_object to create the object.
# * Resources that need to be synced are tracked in the resource_registry,
#   but only go into it when they have their _wgpu_object set (i.e. when
#   the resource actually exists on the GPU).
# * Right before the renderer performs a draw, it queries the registry
#   and calls update_resource on each.


def _update_buffer(buffer):
    wgpu_buffer = buffer._wgpu_object
    assert wgpu_buffer is not None

    # Collect chunks to update
    chunk_descriptions = buffer._gfx_get_chunk_descriptions()
    if not chunk_descriptions:
        return

    # Prepare for data uploads
    device = get_shared().device
    bytes_per_item = buffer.itemsize

    # Upload any pending data
    for chunk_description in chunk_descriptions:
        if len(chunk_description) == 3:
            offset, size, chunk_data = chunk_description
        else:
            offset, size = chunk_description
            chunk_data = buffer._gfx_get_chunk_data(offset, size)
        device.queue.write_buffer(
            wgpu_buffer, bytes_per_item * offset, chunk_data, 0, chunk_data.nbytes
        )
        # todo: investigate using a staging buffer/belt https://github.com/gfx-rs/wgpu-rs/blob/master/src/util/belt.rs


def _update_texture(texture):
    wgpu_texture = texture._wgpu_object
    assert wgpu_texture is not None

    # Collect chunks to update
    chunk_descriptions = texture._gfx_get_chunk_descriptions()
    if not chunk_descriptions:
        return

    # Prepare for data uploads
    device = get_shared().device
    fmt = to_texture_format(texture.format)
    pad_value = 0.0
    extra_bytes = 0
    if texture._wgpu_emulate_rgb:
        _, pad_value, extra_bytes = ALTTEXFORMAT[fmt]

    bytes_per_pixel = (
        texture.nbytes // (texture.size[0] * texture.size[1] * texture.size[2])
        + extra_bytes
    )

    # Upload any pending data
    for chunk_description in chunk_descriptions:
        if len(chunk_description) == 3:
            offset, size, chunk_data = chunk_description
        else:
            offset, size = chunk_description
            chunk_data = texture._gfx_get_chunk_data(offset, size, pad_value)
        bpp = bytes_per_pixel
        if bpp == 0:
            bpp = chunk_data[0, 0, 0].nbytes
        device.queue.write_texture(
            {"texture": wgpu_texture, "origin": offset, "mip_level": 0},
            chunk_data,
            {"bytes_per_row": size[0] * bpp, "rows_per_image": size[1]},
            size,
        )

    if texture.generate_mipmaps:
        generate_texture_mipmaps(texture)


def ensure_wgpu_object(resource):
    """Make sure that the resource (buffer texture, sampler, textureView)
    has a wgpu object attached to it. Returns the native wgpu object.
    """
    if resource._wgpu_object is not None:
        return resource._wgpu_object

    device = get_shared().device

    if isinstance(resource, Buffer):
        if resource.data is not None:
            resource._wgpu_usage |= wgpu.BufferUsage.COPY_DST
        resource._wgpu_object = device.create_buffer(
            size=resource.nbytes, usage=resource._wgpu_usage
        )
        # Mark the resource for sync at the registry (but only if it has pending updates)
        resource._gfx_mark_for_sync()

    elif isinstance(resource, Texture):
        fmt = to_texture_format(resource.format)
        if fmt in ALTTEXFORMAT:
            fmt = ALTTEXFORMAT[fmt][0]
        if resource.colorspace == "tex-srgb" and not fmt.endswith("-srgb"):
            fmt += "-srgb"
        elif resource.colorspace == "srgb" and fmt.endswith("-srgb"):
            logger.warning(
                "Using texture.format 'xx-srgb' AND texture.colorspace 'srgb'."
            )
        if resource.data is not None:
            resource._wgpu_usage |= wgpu.TextureUsage.COPY_DST
        usage = resource._wgpu_usage
        if resource.generate_mipmaps:
            usage |= wgpu.TextureUsage.RENDER_ATTACHMENT  # mipmap needs this
            resource._wgpu_mip_level_count = get_mip_level_count(resource)
        resource._wgpu_object = device.create_texture(
            size=resource.size,
            usage=usage,
            dimension=f"{resource.dim}d",
            format=fmt,
            mip_level_count=resource._wgpu_mip_level_count,
            sample_count=1,  # could allow more to implement msaa
        )
        # Mark the resource for sync at the registry (but only if it has pending updates)
        resource._gfx_mark_for_sync()

    elif isinstance(resource, GfxTextureView):
        wgpu_texture = ensure_wgpu_object(resource.texture)
        if resource.is_default_view:
            resource._wgpu_object = wgpu_texture.create_view()
        else:
            fmt = to_texture_format(resource.format)
            fmt = ALTTEXFORMAT.get(fmt, [fmt])[0]
            resource._wgpu_object = wgpu_texture.create_view(
                format=fmt,
                dimension=resource.view_dim,
                aspect=resource.aspect,
                base_mip_level=resource.mip_range[0],
                mip_level_count=(resource.mip_range[1] - resource.mip_range[0]),
                base_array_layer=resource.layer_range[0],
                array_layer_count=(resource.layer_range[1] - resource.layer_range[0]),
            )

    elif isinstance(resource, GfxSampler):
        amodes = resource.address_mode.replace(",", " ").split() or ["clamp"]
        while len(amodes) < 3:
            amodes.append(amodes[-1])
        filters = resource.filter.replace(",", " ").split() or ["nearest"]
        while len(filters) < 3:
            filters.append(filters[-1])
        ammap = {"clamp": "clamp-to-edge", "mirror": "mirror-repeat"}
        resource._wgpu_object = device.create_sampler(
            address_mode_u=ammap.get(amodes[0], amodes[0]),
            address_mode_v=ammap.get(amodes[1], amodes[1]),
            address_mode_w=ammap.get(amodes[2], amodes[2]),
            mag_filter=filters[0],
            min_filter=filters[1],
            mipmap_filter=filters[2],
            # lod_min_clamp -> use default 0
            # lod_max_clamp -> use default inf
            # compare -> only not-None for comparison samplers!
        )

    return resource._wgpu_object



"""
.. currentmodule:: pygfx.renderers.wgpu

The ``EffectPass`` is the base class for implementing full screen post-processing effects.
This can be subclasses to create custom effects. A few builtin effects are also available.


.. autosummary::
    :toctree: _autosummary/renderers/wgpu/engine/effectpasses
    :template: ../_templates/custom_layout.rst

    EffectPass
    CopyPass
    PPAAPass
    FXAAPass
    DDAAPass
    NoisePass
    DepthPass
    FogPass

"""

import time

import wgpu

from ....utils import array_from_shadertype
from ....utils.color import Color
from ..shader.bindings import BindingDefinitions
from ..shader.templating import apply_templating
from .utils import GpuCache, hash_from_value
from .shared import get_shared
from .binding import Binding


# This cache enables sharing some gpu objects between code that uses
# full-quad shaders. The gain here won't be large in general, but can
# still be worthwhile in situations where many canvases are created,
# such as in Jupyter notebooks, or e.g. when generating screenshots.
# It should also reduce the required work during canvas resizing.
FULL_QUAD_CACHE = GpuCache("full_quad_objects")


FULL_QUAD_SHADER = """
    struct VertexInput {
        @builtin(vertex_index) index: u32,
    };
    struct Varyings {
        @location(0) texCoord: vec2<f32>,
        @builtin(position) position: vec4<f32>,
    };

    @vertex
    fn vs_main(in: VertexInput) -> Varyings {
        var positions = array<vec2<f32>,4>(
            vec2<f32>(0.0, 1.0), vec2<f32>(0.0, 0.0), vec2<f32>(1.0, 1.0), vec2<f32>(1.0, 0.0)
        );
        let pos = positions[in.index];
        var varyings: Varyings;
        varyings.texCoord = vec2<f32>(pos.x, 1.0 - pos.y);
        varyings.position = vec4<f32>(pos * 2.0 - 1.0, 0.0, 1.0);
        return varyings;
    }
"""


def create_full_quad_pipeline(targets, binding_layout, fragment_code):
    """Low-level support for a full-quad pipeline."""
    device = get_shared().device

    # Get bind group layout
    key1 = hash_from_value(binding_layout)
    bind_group_layout = FULL_QUAD_CACHE.get(key1)
    if bind_group_layout is None:
        bind_group_layout = device.create_bind_group_layout(entries=binding_layout)
        FULL_QUAD_CACHE.set(key1, bind_group_layout)

    # Get render pipeline
    key2 = hash_from_value([bind_group_layout, targets, fragment_code])
    render_pipeline = FULL_QUAD_CACHE.get(key2)
    if render_pipeline is None:
        wgsl = FULL_QUAD_SHADER + fragment_code
        shader_module = device.create_shader_module(code=wgsl)

        pipeline_layout = device.create_pipeline_layout(
            bind_group_layouts=[bind_group_layout]
        )

        render_pipeline = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": shader_module,
                "entry_point": "vs_main",
                "buffers": [],
            },
            primitive={
                "topology": wgpu.PrimitiveTopology.triangle_strip,
                "strip_index_format": wgpu.IndexFormat.uint32,
            },
            depth_stencil=None,
            multisample=None,
            fragment={
                "module": shader_module,
                "entry_point": "fs_main",
                "targets": targets,
            },
        )

        # Bind shader module object to the lifetime of the pipeline object
        render_pipeline._gfx_module = shader_module

        FULL_QUAD_CACHE.set(key2, render_pipeline)

    return render_pipeline


# ----- Class hierachy


class FullQuadPass:
    """
    A base class for rendering a full quad, with support for uniforms.
    The only current subclass is the EffectPass. The design is such that it supports
    as many source and target textures as you want, so effects can e.g. have a texture
    as a property.
    """

    uniform_type = dict()

    wgsl = ""

    def __init__(self):
        self._device = get_shared().device

        self._uniform_data = array_from_shadertype(self.uniform_type)
        self._wgpu_buffer = self._device.create_buffer(
            size=self._uniform_data.nbytes,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST,
        )
        bindings_definition = BindingDefinitions()
        bindings_definition.define_binding(
            0, 0, Binding("u_effect", "buffer/uniform", self._uniform_data.dtype)
        )
        self._uniform_binding_definition = bindings_definition.get_code()
        self._uniform_binding_entry = {
            "binding": 0,
            "resource": {
                "buffer": self._wgpu_buffer,
                "offset": 0,
                "size": self._wgpu_buffer.size,
            },
        }

        self._wgpu_sampler = self._device.create_sampler(
            min_filter="linear", mag_filter="linear"
        )
        self._sampler_binding_definition = """
            @group(0) @binding(1)
            var texSampler: sampler;
        """
        self._sampler_binding_entry = {"binding": 1, "resource": self._wgpu_sampler}

        self._render_pipeline = None
        self._render_pipeline_hash = None
        self._template_vars = {}
        self._template_vars_changed = True

    def _set_template_var(self, **kwargs):
        for name, value in kwargs.items():
            if value != self._template_vars.get(name, "stubvaluethatsnotit"):
                self._template_vars_changed = True  # causes a shader recompile
        self._template_vars.update(kwargs)

    def render(
        self,
        command_encoder,
        **texture_views,
    ):
        """Render the source into the destination, applying the postprocessing effect."""

        source_textures = {}
        target_textures = []

        for name, tex in texture_views.items():
            if not isinstance(tex, wgpu.GPUTextureView):
                raise TypeError(f"FullQuadPass expected a texture view, not {tex!r}")
            if name.startswith("target"):
                target_textures.append(tex)
            else:
                source_textures[name] = tex
        if not target_textures:
            raise RuntimeError(
                "FullQuadPass needs at least one target texture (prefixed with 'target')."
            )

        # Make sure we have the render_pipeline
        render_pipeline_hash = (
            tuple(source_textures.keys()),
            tuple(tex.texture.format for tex in target_textures),
        )
        if (
            self._template_vars_changed
            or render_pipeline_hash != self._render_pipeline_hash
        ):
            self._template_vars_changed = False
            self._render_pipeline_hash = render_pipeline_hash
            self._render_pipeline = self._create_pipeline(*render_pipeline_hash)

        # Update uniforms
        self._device.queue.write_buffer(
            self._wgpu_buffer, 0, self._uniform_data, 0, self._uniform_data.nbytes
        )

        # Ready to go!
        return self._render(command_encoder, source_textures.values(), target_textures)

    def _render(self, command_encoder, source_textures, target_textures):
        # Create bind group. This is very light and can be done every time.
        # Chances are we get new views on every call anyway.
        bind_group_entries = [self._uniform_binding_entry, self._sampler_binding_entry]
        for i, tex in enumerate(source_textures, 2):
            bind_group_entries.append({"binding": i, "resource": tex})
        bind_group = self._device.create_bind_group(
            layout=self._render_pipeline.get_bind_group_layout(0),
            entries=bind_group_entries,
        )

        # Create attachments
        color_attachments = []
        for tex in target_textures:
            color_attachments.append(
                {
                    "view": tex,
                    "resolve_target": None,
                    "clear_value": (0, 0, 0, 0),
                    "load_op": wgpu.LoadOp.clear,
                    "store_op": wgpu.StoreOp.store,
                }
            )

        render_pass = command_encoder.begin_render_pass(
            color_attachments=color_attachments,
            depth_stencil_attachment=None,
        )
        render_pass.set_pipeline(self._render_pipeline)
        render_pass.set_bind_group(0, bind_group, [], 0, 99)
        render_pass.draw(4, 1)
        render_pass.end()

    def _create_pipeline(self, source_names, target_formats):
        binding_layout = []
        definitions_code = ""

        # Uniform buffer
        binding_layout.append(
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "buffer": {"type": wgpu.BufferBindingType.uniform},
            }
        )
        definitions_code += self._uniform_binding_definition

        # Sampler
        binding_layout.append(
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "sampler": {"type": wgpu.SamplerBindingType.filtering},
            }
        )
        definitions_code += self._sampler_binding_definition

        # Source textures
        for i, name in enumerate(source_names, 2):
            sample_type = wgpu.TextureSampleType.float
            wgsl_type = "texture_2d<f32>"
            if "depth" in name.lower():
                sample_type = wgpu.TextureSampleType.depth
                wgsl_type = "texture_depth_2d"
            binding_layout.append(
                {
                    "binding": i,
                    "visibility": wgpu.ShaderStage.FRAGMENT,
                    "texture": {
                        "sample_type": sample_type,
                        "view_dimension": wgpu.TextureViewDimension.d2,
                        "multisampled": False,
                    },
                }
            )
            definitions_code += f"""
                @group(0) @binding({i})
                var {name}: {wgsl_type};
            """

        # Render targets
        targets = []
        for format in target_formats:
            targets.append(
                {
                    "format": format,
                    "blend": {
                        "alpha": {
                            "operation": wgpu.BlendOperation.add,
                            "src_factor": wgpu.BlendFactor.one,
                            "dst_factor": wgpu.BlendFactor.zero,
                        },
                        "color": {
                            "operation": wgpu.BlendOperation.add,
                            "src_factor": wgpu.BlendFactor.one,
                            "dst_factor": wgpu.BlendFactor.zero,
                        },
                    },
                }
            )

        wgsl = definitions_code
        wgsl += apply_templating(self.wgsl, **self._template_vars)
        return create_full_quad_pipeline(targets, binding_layout, wgsl)


class EffectPass(FullQuadPass):
    """
    Base class to do post-processing effect passes, converting one image into another.
    """

    # TODO: use templating and keep a dict with template variables. That we we can e.g. scale the support of a Gaussian blur effect properly.

    USES_DEPTH = False
    """Overloadable class attribute to state whether bindings to the depth buffer are needed.
    """

    uniform_type = dict(
        time="f4",
    )
    """Overloadable class attribute that defines the structure of the uniform struct.

    In a subclass you should use something like this::

        uniform_type = dict(
            EffectPass.uniform_type,
            color="4xf4",
            strength="f4",
        )
    """

    wgsl = "EffectPass_needs_to_be_subclassed(and_its_wgsl_attr_overloaded);"
    """Overloadable class attribute that contains the WGSL shading code for the fragment shader.

    Use the following template::

        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {

            // Available variables:
            // colorTex - the texture containing the rendered image, or the previous effect pass.
            // depthTex - the texture containing the renderd depth values.
            // texSampler - a sampler to use for the above.
            // varyings.position - the position in physical pixels (a vec4f).
            // varyings.texCoord - the coordinate in the textures (a vec2f).
            // u_effect.time - the current time in seconds, changes each frame.
            // u_effect.xx - whatever uniforms you added.

            // Calculate the pixel index, e.g. if you want to use textureLoad().
            let texIndex = vec2i(varyings.position.xy);

            // To simply copy the image:
            return textureSample(colorTex, texSampler, varyings.texCoord);
        }
    """

    def __repr__(self):
        return f"<{self.__class__.__name__} at {hex(id(self))}>"

    def render(
        self,
        command_encoder,
        color_tex,
        depth_tex,
        target_tex,
    ):
        """Render the pass using the provided textures.

        If ``USES_DEPTH`` is False, the ``depth_tex`` is ignored (and may be set to None).
        """
        # Set uniforms
        self._uniform_data["time"] = time.perf_counter()

        # Compose kwargs containing the textures. Target textures are prefixed with 'target'
        kwargs = dict(colorTex=color_tex, targetTex=target_tex)
        if self.USES_DEPTH:
            kwargs["depthTex"] = depth_tex

        super().render(command_encoder, **kwargs)


class CopyPass(EffectPass):
    """
    Simple pass that does nothing but copy the texture over, using linear interpolation if the texture size does not match.
    Mostly included for testing.
    """

    wgsl = """
        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
            return textureSample(colorTex, texSampler, varyings.texCoord);
        }
        """


class OutputPass(EffectPass):
    """
    Render from one texture into another, taking size difference into account. Applying gamma on the way.
    """

    # Note: This class is not public, but used internally by the renderer to copy the final result to the target texture. So technically not so much an 'effect'.

    uniform_type = dict(
        EffectPass.uniform_type,
        gamma="f4",
    )

    wgsl = "{$ include 'pygfx.ssaa.wgsl' $}"

    def __init__(self, *, gamma=1.0, filter="mitchell"):
        super().__init__()
        self.gamma = gamma
        self.filter = filter
        self._set_template_var(
            extraKernelSupport=None,  # for testing
            optCorners=True,  # optimization: drop corners in kernels larger than 6x6
            optScale2=True,  # optimization: use 12-tap filters for cubic kernels when scaleFactor == 2
            gamma="u_effect.gamma",  # let gamma use the uniform buffer
        )

    def render(
        self,
        command_encoder,
        color_tex,
        depth_tex,
        target_tex,
    ):
        """Render the (internal) result of the renderer to a texture view."""

        # Get factor from source to destination. A value < 1 means the source is smaller; we're upsampling.
        factor_x = color_tex.size[0] / target_tex.size[0]
        factor_y = color_tex.size[1] / target_tex.size[1]
        factor = (factor_x + factor_y) / 2

        self._set_template_var(scaleFactor=float(factor))

        return super().render(command_encoder, color_tex, depth_tex, target_tex)

    @property
    def gamma(self):
        """The gamma correction to apply."""
        return float(self._uniform_data["gamma"])

    @gamma.setter
    def gamma(self, gamma):
        self._uniform_data["gamma"] = float(gamma)

    @property
    def filter(self):
        """The type of filter to use."""
        return self._template_vars["filter"]

    @filter.setter
    def filter(self, filter):
        filter = filter.lower()
        filters = {
            "nearest",
            "box",
            "linear",
            "tent",
            "disk",
            "mitchell",
            "bspline",
            "catmull",
        }
        if filter not in filters:
            raise ValueError(
                f"Unknown OutputPass filter {filter!r}, must be one of {set(filters)}"
            )
        self._set_template_var(filter=filter)


# ----- Builtin effects

# Creating effects is fun! Some ideas:
#
# BloomPass, FilmPass, GlitchPass, GodRayPass, OutlinePass, AmbientOcclusionPass, FogPass, ...
# GaussXPass, GaussYPass, GaussDxPass, GaussDyPass, SobelPass, LaplacianPass, SharpeningPass, ...
# Color grading, color conversions, depth of field, patterns, pixelize, tone mapping, texture overlay, ...
#
# See https://github.com/pmndrs/postprocessing and ThreeJS code for implementations.


class PPAAPass(EffectPass):
    """Base class for post-processing anti-aliasing to help the renderer detect these."""


class FXAAPass(PPAAPass):
    """An effect pass implementing Fast approximate anti-aliasing.

    FXAA is a well known method for post-processing antialiasing.
    This is version 3.11.
    """

    wgsl = "{$ include 'pygfx.fxaa3.wgsl' $}"


class DDAAPass(PPAAPass):
    """An effect pass implementing Directional Diffusion anti-aliasing.

    DDAA produces better results than FXAA for near-diagonal lines, at the same performance.
    It estimates the direction of the edge, and then diffuses (i.e. smoothes) in that direction.
    For near-horizontal and near-vertical a technique similar to FXAA is used.
    """

    wgsl = "{$ include 'pygfx.ddaa2.wgsl' $}"

    def __init__(self, *, max_edge_iters=5):
        super().__init__()
        self.max_edge_iters = max_edge_iters

    @property
    def max_edge_iters(self):
        """The maximum number of iters (of 3 samples) to search along an edge.

        Default 5 (i.e. search 15 pixels along an edge). Set higger for prettier
        edges, or lower for more performance.
        """
        return len(self._template_vars["EDGE_STEP_LIST"])

    @max_edge_iters.setter
    def max_edge_iters(self, max_edge_iters):
        # The default EDGE_STEP_LIST is [3, 3, 3, 3, 3]
        edge_step_list = [3] * int(max_edge_iters)
        self._set_template_var(EDGE_STEP_LIST=edge_step_list)


class NoisePass(EffectPass):
    """An effect pass that adds noise."""

    uniform_type = dict(
        EffectPass.uniform_type,
        noise="f4",
    )

    wgsl = """
        {$ include 'pygfx.noise.wgsl' $}

        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
            let texCoord = varyings.texCoord;
            let texIndex = vec2i(varyings.position.xy);
            let noise = random(texCoord.x * texCoord.y * u_effect.time);
            let color = textureLoad(colorTex, texIndex, 0);
            return color + noise * u_effect.noise;
        }
    """

    def __init__(self, noise=0.1):
        super().__init__()
        self.noise = noise

    @property
    def noise(self):
        """The strength of the noise."""
        return float(self._uniform_data["noise"])

    @noise.setter
    def noise(self, noise):
        self._uniform_data["noise"] = float(noise)


class DepthPass(EffectPass):
    """An effect that simply renders the depth. Mostly for debugging purposes."""

    USES_DEPTH = True

    wgsl = """
        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
            let texIndex = vec2i(varyings.position.xy);
            let depth = textureLoad(depthTex, texIndex, 0);
            return vec4f(depth, depth, depth, 1.0);
        }
    """


class FogPass(EffectPass):
    """An effect pass that adds fog to the full image, using the depth buffer."""

    USES_DEPTH = True

    uniform_type = dict(
        EffectPass.uniform_type,
        color="4xf4",
        power="f4",
    )

    wgsl = """
        @fragment
        fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
            let texIndex = vec2i(varyings.position.xy);
            let raw_depth = textureLoad(depthTex, texIndex, 0);
            let depth = pow(raw_depth, u_effect.power);

            let color = textureLoad(colorTex, texIndex, 0);
            let depth_color = u_effect.color;

            return mix(color, depth_color, depth);
        }
    """

    def __init__(self, color="#fff", power=1.0):
        super().__init__()
        self.color = color
        self.power = power

    @property
    def color(self):
        """The color of the fog."""
        return Color(self._uniform_data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self._uniform_data["color"] = color

    @property
    def power(self):
        """The power to apply to the depth value. Using a value < 1 brings the range more towards the camera."""
        return float(self._uniform_data["power"])

    @power.setter
    def power(self, power):
        self._uniform_data["power"] = float(power)



import time


class Clock:
    """A simple clock for keeping track of time"""

    def __init__(self, auto_start=True):
        self._auto_start = auto_start

        self._start_time = None
        self._last_time = None
        self._elapsed_time = 0
        self._running = False

    def start(self):
        self._start_time = self._last_time = time.perf_counter()
        self._elapsed_time = 0
        self._running = True

    def stop(self):
        self._running = False

    def get_elapsed_time(self):
        self.get_delta()
        return self._elapsed_time

    def get_delta(self):
        diff = 0

        if self._auto_start and not self._running:
            self.start()
            return 0

        if self._running:
            now = time.perf_counter()
            diff = now - self._last_time
            self._last_time = now

            self._elapsed_time += diff

        return diff



"""Predefined Colormaps.

.. currentmodule:: pygfx.utils.cm

.. autosummary::

    create_colormap
    gray
    viridis
    plasma
    inferno
    magma
    cividis
    cool
    hot
    bone
    copper
    pink
    spring
    summer
    autumn
    winter
    jet

"""

import numpy as np

from ..resources import Texture, TextureMap
from ._cmdata_mpl import (
    _viridis_data,
    _plasma_data,
    _inferno_data,
    _magma_data,
    _cividis_data,
)


_data = {
    # Simple maps
    "gray": [(0, 0, 0, 0), (1, 1, 1, 1)],
    "cool": [(0, 1, 1), (1, 0, 1)],
    "hot": [(0, 0, 0), (1, 0, 0), (1, 1, 0), (1, 1, 1)],
    "bone": [(0, 0, 0), (0.333, 0.333, 0.444), (0.666, 0.777, 0.777), (1, 1, 1)],
    "copper": [(0, 0, 0), (1, 0.7, 0.5)],
    "pink": [(0.1, 0, 0), (0.75, 0.5, 0.5), (0.9, 0.9, 0.7), (1, 1, 1)],
    "spring": [(1, 0, 1), (1, 1, 0)],
    "summer": [(0, 0.5, 0.4), (1, 1, 0.4)],
    "autumn": [(1, 0, 0), (1, 1, 0)],
    "winter": [(0, 0, 1), (0, 1, 0.5)],
    "jet": [
        (0, 0, 0.5),
        (0, 0, 1),
        (0, 0.5, 1),
        (0, 1, 1),
        (0.5, 1, 0.5),
        (1, 1, 0),
        (1, 0.5, 0),
        (1, 0, 0),
        (0.5, 0, 0),
    ],
    # Larger maps
    "viridis": _viridis_data,
    "plasma": _plasma_data,
    "inferno": _inferno_data,
    "magma": _magma_data,
    "cividis": _cividis_data,
}

__all__ = sorted([*_data.keys(), "create_colormap"])


def create_colormap(input, n=256):
    """Create a colormap from given data.

    Returns a TextureMap with (at least) n colors.

    If a list or array-like object is given, the values are interpolated to create the map.
    E.g. ``create_colormap([(0, 0, 0), (1, 1, 1)])`` creates the gray colormap.

    If a dict is given, it must have at least one field of "r", "g", "b", "a".
    Each value is a sequence of 2 values (i.e. an Nx2 array-like), where the
    first value specifies the index (0..1), and the second the value for that color.
    """

    if isinstance(input, dict):
        # Init data, alpha 1
        colormap_data = np.zeros((n, 4), "f4")
        colormap_data[:, 3] = 1.0
        # For each channel ...
        for channel_index in range(4):
            channe_name = "rgba"[channel_index]
            try:
                values = input[channe_name]
            except KeyError:
                continue  # this color not in dict
            # Get value list and check
            values = np.asarray(values)
            if not (values.ndim == 2 and values.shape[1] == 2):
                raise ValueError(
                    "The values in the dict to convert to a colormap must be Nx2."
                )
            # Init interpolation
            data = np.zeros((len(values),), dtype="f4")
            x = np.linspace(0.0, 1.0, n)
            xp = np.zeros((len(values),), dtype="f4")
            # Insert values
            for i, el in enumerate(values):
                xp[i] = el[0]
                data[i] = el[1]
            # Interpolate
            colormap_data[:, channel_index] = np.interp(x, xp, data)

    else:
        # Input is list or array-like
        data = np.array(input, dtype="f4")
        if not (data.ndim == 2 and data.shape[1] in [3, 4]):
            raise ValueError("Colormap entries must have 3 or 4 elements.")
        if data.shape[1] == 3:
            data = np.hstack([data, np.ones((data.shape[0], 1), "f4")])

        # Apply interpolation (if required)
        if data.shape[0] == n:
            colormap_data = data
        else:
            x = np.linspace(0.0, 1.0, n)
            xp = np.linspace(0.0, 1.0, data.shape[0])
            colormap_data = np.zeros((n, 4), "f4")
            for i in range(4):
                colormap_data[:, i] = np.interp(x, xp, data[:, i])

    return TextureMap(Texture(colormap_data, dim=1), filter="nearest", wrap="clamp")


def __dir__():
    return __all__


def __getattr__(name):
    if name not in _data:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
    colormap = create_colormap(_data[name])
    globals()[name] = colormap  # prevent creating it again
    return colormap



"""Objects to view a scene.

.. currentmodule:: pygfx.cameras

.. autosummary::
    :toctree: cameras/
    :template: ../_templates/custom_layout.rst

    Camera
    NDCCamera
    ScreenCoordsCamera
    PerspectiveCamera
    OrthographicCamera

"""

from ._base import Camera, NDCCamera, ScreenCoordsCamera
from ._perspective import PerspectiveCamera
from ._orthographic import OrthographicCamera

__all__ = [
    "Camera",
    "NDCCamera",
    "OrthographicCamera",
    "PerspectiveCamera",
    "ScreenCoordsCamera",
]



import numpy as np
import wgpu

from ....resources._texture import Texture

from .utils import GfxTextureView, GpuCache
from .shared import get_shared


# This cache enables re-using gpu pipelines for calculating mipmaps, so
# that these don't have to be created on each draw, which would make
# things very slow. The number of pipelines in the cache won't be large
# since there are only so many texture formats, but it seems cleaner
# to use our cache mechanism than to store them globally.
MIPMAP_CACHE = GpuCache("mipmap_pipelines")

mipmap_sampler = None

mipmap_vertex_source = """

    struct VarysStruct {
        @builtin( position ) Position: vec4<f32>,
        @location( 0 ) vTex : vec2<f32>
    };
    @vertex
    fn main( @builtin( vertex_index ) vertexIndex : u32 ) -> VarysStruct {
        var Varys : VarysStruct;
        var pos = array< vec2<f32>, 4 >(
            vec2<f32>( -1.0,  1.0 ),
            vec2<f32>(  1.0,  1.0 ),
            vec2<f32>( -1.0, -1.0 ),
            vec2<f32>(  1.0, -1.0 )
        );
        var tex = array< vec2<f32>, 4 >(
            vec2<f32>( 0.0, 0.0 ),
            vec2<f32>( 1.0, 0.0 ),
            vec2<f32>( 0.0, 1.0 ),
            vec2<f32>( 1.0, 1.0 )
        );
        Varys.vTex = tex[ vertexIndex ];
        Varys.Position = vec4<f32>( pos[ vertexIndex ], 0.0, 1.0 );
        return Varys;
    }
"""

mipmap_fragment_source = """

    @group( 0 ) @binding( 0 )
    var imgSampler : sampler;
    @group( 0 ) @binding( 1 )
    var img : texture_2d<f32>;
    @fragment
    fn main( @location( 0 ) vTex : vec2<f32> ) -> @location( 0 ) vec4<f32> {
        return textureSample( img, imgSampler, vTex );
    }
"""


def get_mip_level_count(texture):
    """Get the number of mipmap levels from the texture size."""
    if not isinstance(texture, Texture):
        raise TypeError("Expecting a Texture object.")
    width, height, _ = texture.size
    return int(np.floor(np.log2(max(width, height))) + 1)


def generate_texture_mipmaps(target):
    """Generate mipmaps for the given target. The target can be a
    Texture or GfxTextureView and can be a 2D texture as well as a cube
    texture.
    """

    # If this looks like a cube or stack, generate mipmaps for each individual layer
    if isinstance(target, Texture) and target.dim == 2 and target.size[2] > 1:
        for i in range(target.size[2]):
            generate_texture_mipmaps(GfxTextureView(target, layer_range=(i, i + 1)))
        return

    if isinstance(target, Texture):
        texture = target
        layer = 0
    elif isinstance(target, GfxTextureView):
        view, texture = target, target.texture
        layer = view.layer_range[0]

    generate_mipmaps(texture, layer)


def generate_mipmaps(texture, base_array_layer):
    device = get_shared().device

    pipeline = get_mipmap_pipeline(device, texture)

    command_encoder: "wgpu.GPUCommandEncoder" = device.create_command_encoder()
    bind_group_layout = pipeline.get_bind_group_layout(0)

    dst_size = texture.size[:2]
    prev_view = None

    for i in range(1, texture._wgpu_mip_level_count):
        dst_size = dst_size[0] // 2, dst_size[1] // 2
        bind_group, prev_view = get_bind_group(
            device, texture, bind_group_layout, base_array_layer, i, prev_view
        )
        pass_encoder = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": prev_view,
                    "load_op": wgpu.LoadOp.clear,
                    "store_op": wgpu.StoreOp.store,
                    "clear_value": [0, 0, 0, 0],
                }
            ]
        )

        pass_encoder.set_pipeline(pipeline)
        pass_encoder.set_bind_group(0, bind_group)
        pass_encoder.draw(4, 1, 0, 0)
        pass_encoder.end()

    device.queue.submit([command_encoder.finish()])


def get_bind_group(
    device, texture, bind_group_layout, base_array_layer, mip_level, src_view
):
    # Get the bind group that includes the two subsequent texture views.
    # We cache this on the texture object to avoid re-creating the views
    # and bind-group each time, which results in a significant increase
    # in performance.
    #
    # Note, however, that this does mean we hold on to GPU resources,
    # which may be a bit of a waste if the mipmaps are only created once!
    # So for now we're assuming that mipmaps are generated often.

    key = f"_gfx_bind_group_{base_array_layer}_{mip_level}"
    bind_group = getattr(texture, key, None)
    dst_view = texture._wgpu_object.create_view(
        base_mip_level=mip_level,
        mip_level_count=1,
        dimension="2d",
        base_array_layer=base_array_layer,
    )
    if bind_group is None:
        if src_view is None:
            src_view = texture._wgpu_object.create_view(
                base_mip_level=mip_level - 1,
                mip_level_count=1,
                dimension="2d",
                base_array_layer=base_array_layer,
            )
        global mipmap_sampler
        if mipmap_sampler is None:
            mipmap_sampler = device.create_sampler(min_filter="linear")

        bind_group = device.create_bind_group(
            layout=bind_group_layout,
            entries=[
                {"binding": 0, "resource": mipmap_sampler},
                {"binding": 1, "resource": src_view},
            ],
        )
        setattr(texture, key, bind_group)

    return bind_group, dst_view


def get_mipmap_pipeline(device, texture):
    format = texture._wgpu_object.format
    pipeline = MIPMAP_CACHE.get(format)

    if pipeline is None:
        vertex_module = device.create_shader_module(code=mipmap_vertex_source)

        frag_module = device.create_shader_module(code=mipmap_fragment_source)

        pipeline = device.create_render_pipeline(
            layout="auto",
            vertex={
                "module": vertex_module,
                "entry_point": "main",
            },
            fragment={
                "module": frag_module,
                "entry_point": "main",
                "targets": [{"format": format}],
            },
            primitive={
                "topology": "triangle-strip",
                "strip_index_format": "uint32",
            },
        )

        MIPMAP_CACHE.set(format, pipeline)

    # Store a ref of the pipeline, the cache uses weak refs.
    # I.e. the pipeline for this format is in the cache for as long as
    # any textures that use it are alive.
    texture._wgpu_mipmap_pipeline = pipeline

    return pipeline



"""
The engine of the wgpu renderer.

This is the only place in pygfx where we make wgpu function calls.
(Although we may use enums or flags elsewhere.)

----

Below is the high level model in which the visualization is described.
Let's call this level 1. There is the obvious world object, material,
and geometry. There also is global data where the renderer stores camera
transform, and canvas sizes (in a single global uniform). And then there
is the things defined by the renderstate (i.e. scene) like lights and
shadows. Renderer specific state is also part of the renderstate.

                           Uniform-Buffer
                              ▲
       Uniform-Buffer         │
              ▲           ┌───┴────┐      ┌──────────────┐    ┌─────────────┐
              │        ┌──┤Material│      │Renderstate:  │    │Global:      │
        ┌─────┴─────┐  │  └────────┘      │- light slots │    │- glyph atlas│
        │WorldObject├──┤                  │- shadow maps │    │- stdinfo    │
        └───────────┘  │  ┌────────┐      │- clip planes?│    │             │
                       └──┤Geometry│      └──────┬───────┘    └──────┬──────┘
                          └───┬────┘             │                   │
                              │                  │                   ▼
                              ▼                  ▼              Uniform-Buffer
                            Buffers &        Affects wgsl         and Texture
                            Textures         and pipeline

From all this stuff, we create an intermediate representation. Let's
call this level 2. This information is created by the BaseShader
corresponding to a certain material. Note that this object is agnostic
about the renderstate.

    ┌─────────────┐  ┌──────────────────┐  ┌───────────────┐  ┌─────────────┐
    │ WGSL source │  │ Resources:       │  │ pipeline info │  │ render info │
    │ (templated) │  │ - index buffer   │  │ (dict)        │  │ (dict)      │
    │             │  │ - vertex buffers │  │               │  │             │
    │             │  │ - bindings       │  │               │  │             │
    └─────────────┘  └──────────────────┘  └───────────────┘  └─────────────┘

The objects above are not wgpu-specific, but make it easy to create
wgpu-specific objects that we see below (level 3). This step is
performed by the PipelineContainer. Some of these objects match 1-on-1
to the world object, but others (shader module and pipeline) are
specific to the renderstate. And then there are multiple blend-passes
to create as well. The renderstate also plays a role in "finalizing"
the final renderstate-specific objects.

                     ┌────────────────┐  ┌──────────────────────────┐  ┌───────────────────────┐
   RenderContext ───►│WGSL source     │  │BindGroupLayoutDescriptors│  │VertexLayoutDescriptors│
   specific          └───────┬────────┘  └──────────┬───────────────┘  └──────────┬────────────┘
   templating                │                      │                             │
                     ┌───────▼────────┐  ┌──────────▼─────────┐                   │
                     │WgpuShaderModule│  │WgpuBindGroupLayouts│                   │
                     └───────┬────────┘  └──────────┬─────────┘                   │
                             │                      │                             │
                             └──────────────────┐   │  ┌──────────────────────────┘
                                                │   │  │
                                            ┌───▼───▼──▼───┐         ┌─────────────┐
                                            │ WgpuPipeline ◄─────────┤pipeline info│
                                            └───────┬──────┘         └─────────────┘
                                                    │
                      Uniform-Buffers               │
                          Buffers   ────────────┐   │
                          Textures              │   │
                             │                  │   │
                             │                  │   │
                      ┌──────▼───────┐      ┌───▼───▼─────────┐      ┌───────────┐
                      │WgpuBindGroups├──────► Dispatch (draw) ◄──────┤render info│
                      └──────────────┘      └─────────────────┘      └───────────┘

## Reacting to changes

To keep a PipelineContainer up-to-date we need to do two things. First
any changes in the world object and/or material must be applied. This
is done via the Trackable classes, and affects the second level. Next,
some of these changes invalidate the objects in level 3, so we need to
detect that as well.

Before each draw, all buffers and textures that have pending changes are
flushed.

## Caching

Objects are stored in a `WeakAssociativeContainer`, effectively binding
their lifetime to both the world object and renderstate object.

We also re-use wgpu objects like shadermodules. If there are a lot of
objects in a scene, its likely that many of these have the same material.

(Figures created with https://asciiflow.com/)
"""



import numpy as np

from ....utils import array_from_shadertype
from ....resources import Buffer

from ..engine.utils import (
    to_vertex_format,
    to_texture_format,
    generate_uniform_struct,
)


class BindingDefinitions:
    """Track definitions of bindings."""

    def __init__(self):
        self._typedefs = {}
        self._binding_codes = {}
        self._uniform_struct_names = {}  # dtype -> name

    def clear(self):
        self._typedefs.clear()
        self._binding_codes.clear()
        self._uniform_struct_names.clear()

    def get_code(self):
        """Get the wgsl source code for the collected bindings."""
        code = ""
        code += "\n".join(self._typedefs.values())
        code += "\n"
        code += "\n".join(self._binding_codes.values())
        return code

    def define_binding(self, bindgroup, index, binding):
        """Define a uniform, buffer, sampler, or texture. The produced wgsl
        will be part of the code returned by ``get_definitions()``. The binding
        must be a Binding object.
        """
        if binding.type == "buffer/uniform":
            self._define_uniform(bindgroup, index, binding)
        elif binding.type.startswith("buffer"):
            self._define_buffer(bindgroup, index, binding)
        elif binding.type.startswith("sampler"):
            self._define_sampler(bindgroup, index, binding)
        elif binding.type.startswith("texture"):
            self._define_texture(bindgroup, index, binding)
        else:
            raise RuntimeError(
                f"Unknown binding {binding.name} with type {binding.type}"
            )

    def _define_uniform(self, bindgroup, index, binding):
        resource = binding.resource
        if isinstance(resource, dict):
            dtype_struct = array_from_shadertype(resource).dtype
        elif isinstance(resource, Buffer):
            if resource.data.dtype.names is None:
                raise TypeError("define_uniform() needs a structured dtype")
            dtype_struct = resource.data.dtype
        elif isinstance(resource, np.dtype):
            if resource.names is None:
                raise TypeError("define_uniform() needs a structured dtype")
            dtype_struct = resource
        else:
            raise TypeError(f"Unsupported struct type {resource.__class__.__name__}")

        # Get struct name
        struct_hash = str(dtype_struct)

        try:
            structname = self._uniform_struct_names[struct_hash]
            if binding.structname is not None:
                # Do we need to ensure that a dtype corresponds to only one struct?
                assert structname == binding.structname, (
                    "dtype[{struct_hash}] has been defined as struct[{structname}]"
                )
        except KeyError:
            # sometimes, we need a meaningful alias for the struct name.
            if binding.structname is not None:
                structname = binding.structname
                assert structname not in self._uniform_struct_names.values(), (
                    "structname has been used for another dtype"
                )
            else:
                # auto generate struct name
                structname = f"Struct_u_{len(self._uniform_struct_names) + 1}"

            self._uniform_struct_names[struct_hash] = structname

        if structname not in self._typedefs:
            struct_code = generate_uniform_struct(dtype_struct, structname)
            self._typedefs[structname] = struct_code

        uniform_type_name = (
            f"array<{structname}, {binding.resource.data.shape[0]}>"  # array of struct
            if isinstance(resource, Buffer) and resource.data.shape  # Buffer.items > 1
            else structname
        )

        code = f"""
        @group({bindgroup}) @binding({index})
        var<uniform> {binding.name}: {uniform_type_name};
        """.rstrip()
        self._binding_codes[binding.name] = code

    def _define_buffer(self, bindgroup, index, binding):
        # Get format, and split in the scalar part and the number of channels
        fmt = to_vertex_format(binding.resource.format)
        if "x" in fmt:
            fmt_scalar, _, nchannels = fmt.partition("x")
            nchannels = int(nchannels)
        else:
            fmt_scalar = fmt
            nchannels = 1

        # Define scalar type: i32, u32 or f32
        # Since the stride must be a multiple of 4 for storage buffers,
        # the supported types is limited until we support structured numpy arrays.
        scalar_type = (
            fmt_scalar.replace("float", "f").replace("uint", "u").replace("sint", "i")
        )
        if not scalar_type.endswith("32"):
            raise ValueError(
                f"Buffer format {fmt} not supported, format must have a stride of 4 bytes: i4, u4 of f4."
            )

        # Define the element types. The element_type2 is the actual type.
        # Because for storage buffers a vec3 has an alignment of 16, we have to
        # be creative for vec3: we bind the buffer as if it was 1D, and convert
        # in the accessor function.
        if nchannels == 1:
            element_type1 = element_type2 = scalar_type
            stride = 4
        elif nchannels == 3:
            element_type1 = scalar_type
            element_type2 = f"vec{nchannels}<{scalar_type}>"
            stride = 4
        else:
            element_type1 = element_type2 = f"vec{nchannels}<{scalar_type}>"
            stride = 4 * nchannels

        stride  # noqa: B018 - not actually used anymore in wgsl?

        # Produce the binding code and accessor function
        type_modifier = "read" if "read_only" in binding.type else "read_write"
        code = f"""
        @group({bindgroup}) @binding({index})
        var<storage, {type_modifier}> {binding.name}: array<{element_type1}>;
        fn load_{binding.name} (i: i32) -> {element_type2} {{
        """.rstrip()
        if element_type1 == element_type2:
            code += f" return {binding.name}[i];"
        elif nchannels == 2:
            code += f" return {element_type2}( {binding.name}[i * 2], {binding.name}[i * 2 + 1] );"
        elif nchannels == 3:
            code += f" return {element_type2}( {binding.name}[i * 3], {binding.name}[i * 3 + 1], {binding.name}[i * 3 + 2] );"
        else:  # nchannels == 4
            code += f" return {element_type2}( {binding.name}[i * 4], {binding.name}[i * 4 + 1], {binding.name}[i * 4 + 2], {binding.name}[i * 4 + 3] );"
        code += " }"
        self._binding_codes[binding.name] = code

    def _define_sampler(self, bindgroup, index, binding):
        code = f"""
        @group({bindgroup}) @binding({index})
        var {binding.name}: sampler;
        """.rstrip()
        self._binding_codes[binding.name] = code

    def _define_texture(self, bindgroup, index, binding):
        texture = binding.resource  # GfxTextureView
        format = to_texture_format(texture.format)
        if "norm" in format or "float" in format:
            format = "f32"
        elif "uint" in format:
            format = "u32"
        else:
            format = "i32"

        view_dim = texture.view_dim
        view_dim = view_dim.replace("-", "_")  # 2d-array -> 2d_array

        code = f"""
        @group({bindgroup}) @binding({index})
        var {binding.name}: texture_{view_dim}<{format}>;
        """.rstrip()
        self._binding_codes[binding.name] = code



from ..resources import Texture, TextureMap
from ._base import Material
from ..utils import Color, assert_type


class VolumeBasicMaterial(Material):
    """Basic volume material.

    Parameters
    ----------
    clim : tuple
        The contrast limits to scale the data values with. Default (0, 1).
    map : TextureMap | Texture
        The colormap to turn the voxel values into their final color.
    gamma : float
        The gamma correction to apply to the image data. Default 1.
    interpolation : str
        The method to interpolate the image data. Either 'nearest' or 'linear'. Default 'linear'.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    """

    uniform_type = dict(
        Material.uniform_type,
        clim="2xf4",
        gamma="f4",
    )

    def __init__(
        self,
        clim=None,
        map=None,
        gamma=1.0,
        interpolation="linear",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.map = map
        self.clim = clim
        self.gamma = gamma
        # Note: the default volume interpolation is 'linear' while it's nearest
        # for images. The ability to spot the individual voxels simply results in
        # poor visual quality.
        self.interpolation = interpolation

    @property
    def map(self):
        """The colormap to turn the voxel values into their final color.
        If not given or None, the values themselves represents the color.
        The dimensionality of the map can be 1D, 2D or 3D, but should match the
        number of channels in the volume.
        """
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.map = map

    @property
    def clim(self):
        """The contrast limits to scale the data values with. Default (0, 1)."""
        v1, v2 = self.uniform_buffer.data["clim"]
        return float(v1), float(v2)

    @clim.setter
    def clim(self, clim):
        # Check and store given clim
        if clim is None:
            clim = 0, 1
        clim = float(clim[0]), float(clim[1])
        # Update uniform data
        self.uniform_buffer.data["clim"] = clim
        self.uniform_buffer.update_full()

    @property
    def gamma(self):
        """The gamma correction to apply to the image data. Default 1."""
        return self.uniform_buffer.data["gamma"]

    @gamma.setter
    def gamma(self, value):
        value = float(value)
        if value <= 0:
            raise ValueError("gamma must be greater than 0")
        self.uniform_buffer.data["gamma"] = float(value)
        self.uniform_buffer.update_full()

    @property
    def interpolation(self):
        """The method to interpolate the image data. Either 'nearest' or 'linear'."""
        return self._store.interpolation

    @interpolation.setter
    def interpolation(self, value):
        assert value in ("nearest", "linear")
        self._store.interpolation = value


class VolumeSliceMaterial(VolumeBasicMaterial):
    """A material for rendering a slice through a 3D texture at the surface of a mesh.
    This material is not affected by lights.
    """

    uniform_type = dict(
        VolumeBasicMaterial.uniform_type,
        plane="4xf4",
    )

    def __init__(self, plane=(0, 0, 1, 0), **kwargs):
        super().__init__(**kwargs)
        self.plane = plane

    @property
    def plane(self):
        """The plane to slice at, represented with 4 floats ``(a, b, c, d)``,
        which make up the equation: ``ax + by + cz + d = 0`` The plane
        definition applies to the world space (of the scene).
        """
        return self.uniform_buffer.data["plane"]

    @plane.setter
    def plane(self, plane):
        self.uniform_buffer.data["plane"] = plane
        self.uniform_buffer.update_full()


class VolumeRayMaterial(VolumeBasicMaterial):
    """The base class for materials that render volumes using raycasting."""

    render_mode = None  # to be set by the subclasses


class VolumeMipMaterial(VolumeRayMaterial):
    """A material rendering a volume using MIP rendering."""

    render_mode = "mip"


class VolumeMinipMaterial(VolumeRayMaterial):
    """A material rendering a volume using MinIP rendering.

    This material renders the minimum intensity along the view ray.
    """

    render_mode = "minip"


class VolumeIsoMaterial(VolumeRayMaterial):
    """A material rendering a volume using isosurface rendering.

    Parameters
    ----------
    threshold : float
        The threshold texture value at which the surface is rendered.
        The default value is 0.5.
    step_size : float
        The size of the initial ray marching step for the initial surface finding.
        Smaller values will result in more accurate surfaces but slower rendering.
        Default value is 1.0.
    substep_size : float
        The size of the raymarching step for the refined surface finding.
        Smaller values will result in more accurate surfaces but slower rendering.
        Default value is 0.1.
    emissive : Color
        The emissive color of the surface. I.e. the color that the object emits
        even when not lit by a light source. This color is added to the final
        color and unaffected by lighting. The alpha channel is ignored.
        Default value is (0, 0, 0, 1).
    shininess : int
        How shiny the specular highlight is; a higher value gives a sharper
        highlight. Default value is 30.
    """

    render_mode = "iso"
    uniform_type = dict(
        VolumeBasicMaterial.uniform_type,
        threshold="f4",
        emissive_color="4xf4",
        shininess="f4",
        step_size="f4",
        substep_size="f4",
    )

    def __init__(
        self,
        threshold: float = 0.5,
        step_size: float = 1.0,
        substep_size: float = 0.1,
        emissive: Color = "#000",
        shininess: float = 30,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.threshold = threshold
        self.emissive = emissive
        self.shininess = shininess
        self.step_size = step_size
        self.substep_size = substep_size

    @property
    def threshold(self) -> float:
        """The threshold texture value at which the surface is rendered."""
        return self.uniform_buffer.data["threshold"]

    @threshold.setter
    def threshold(self, threshold: float) -> None:
        self.uniform_buffer.data["threshold"] = float(threshold)
        self.uniform_buffer.update_full()

    @property
    def step_size(self) -> float:
        """The size of the initial ray marching step for finding the surface.

        Smaller values will result in more accurate surfaces but slower rendering.
        """
        return self.uniform_buffer.data["step_size"]

    @step_size.setter
    def step_size(self, size: float) -> None:
        self.uniform_buffer.data["step_size"] = float(size)
        self.uniform_buffer.update_full()

    @property
    def substep_size(self) -> float:
        """The size of the raymarching step for the refined surface finding.

        Smaller values will result in more accurate surfaces but slower rendering.
        """
        return self.uniform_buffer.data["substep_size"]

    @substep_size.setter
    def substep_size(self, size: float) -> None:
        self.uniform_buffer.data["substep_size"] = float(size)
        self.uniform_buffer.update_full()

    @property
    def emissive(self):
        """The emissive (light) color of the surface.
        This color is added to the final color and is unaffected by lighting.
        The alpha channel of this color is ignored.
        """
        return Color(self.uniform_buffer.data["emissive_color"])

    @emissive.setter
    def emissive(self, color):
        color = Color(color)
        self.uniform_buffer.data["emissive_color"] = color
        self.uniform_buffer.update_full()

    @property
    def shininess(self):
        """How shiny the specular highlight is; a higher value gives a sharper highlight.
        Default is 30.
        """
        return float(self.uniform_buffer.data["shininess"])

    @shininess.setter
    def shininess(self, value):
        self.uniform_buffer.data["shininess"] = float(value)
        self.uniform_buffer.update_full()



import numpy as np


def merge(groups):
    positions = np.concatenate([g[0] for g in groups])
    normals = np.concatenate([g[1] for g in groups])
    texcoords = np.concatenate([g[2] for g in groups])
    indices = np.concatenate([g[3] for g in groups])
    i = 0
    j = 0
    for g in groups:
        indices[i:] += j
        # advance cursor to start of next group index
        i += len(g[3])
        j = len(g[0])
    return positions, normals, texcoords, indices



"""
Utilities to load gltf/glb files, completely compatible with the glTF 2.0 specification.

References:
https://raw.githubusercontent.com/KhronosGroup/glTF/main/specification/2.0/figures/gltfOverview-2.0.0d.png
https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html

Experimental, not yet fully implemented.
"""

import pygfx as gfx
import numpy as np
import pylinalg as la

from importlib.util import find_spec
from functools import lru_cache


def load_gltf(path, quiet=False, remote_ok=True):
    """
    Load a gltf file and return the content.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    quiet : bool
        Whether to suppress the warning messages.
        Default is False.
    remote_ok : bool
        Whether to allow loading from URLs.
        Default is True.

    Returns:
    ----------
    gltf : object
        The gltf object which contains the following attributes:
        * `scenes`: [gfx.Group]
        * `scene`: gfx.Group or None
        * `cameras`: [gfx.Camera] or None
        * `animations`: [gfx.Animation] or None
    """
    return _GLTF().load(path, quiet, remote_ok)


async def load_gltf_async(path, quiet=False, remote_ok=True):
    """
    Load a gltf file and return the content asynchronously.

    It is recommended to use this function when loading from URLs.
    especially when the gltf model has multiple resources.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    quiet : bool
        Whether to suppress the warning messages.
        Default is False.
    remote_ok : bool
        Whether to allow loading from URLs.
        Default is True.

    Returns:
    ----------
    gltf : object
        The gltf object which contains the following attributes:
        * `scenes`: [gfx.Group]
        * `scene`: gfx.Group or None
        * `cameras`: [gfx.Camera] or None
        * `animations`: [gfx.Animation] or None
    """
    return await _GLTF().load_async(path, quiet, remote_ok)


def load_gltf_mesh(path, materials=True, quiet=False, remote_ok=True):
    """
    Load meshes from a gltf file, without skeletons, and no transformations applied.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    materials : bool
        Whether to load materials.
        Default is True.
    quiet : bool
        Whether to suppress the warning messages.
        Default is False.
    remote_ok : bool
        Whether to allow loading from URLs.
        Default is True.

    Returns:
    ----------
    meshes : list
        A list of pygfx.Meshes.
    """
    return _GLTF().load_mesh(path, quiet, materials=materials, remote_ok=remote_ok)


async def load_gltf_mesh_async(path, materials=True, quiet=False, remote_ok=True):
    """
    Load meshes from a gltf file asynchronously, without skeletons, and no transformations applied.

    It is recommended to use this function when loading from URLs.
    especially when the gltf model has multiple resources.

    This function requires the gltflib library.

    Parameters:
    ----------
    path : str
        The path to the gltf file.
    materials : bool
        Whether to load materials.
        Default is True.
    quiet : bool
        Whether to suppress the warning messages.
        Default is False.
    remote_ok : bool
        Whether to allow loading from URLs.
        Default is True.

    Returns:
    ----------
    meshes : list
        A list of pygfx.Meshes.
    """
    return await _GLTF().load_mesh_async(
        path, quiet, materials=materials, remote_ok=remote_ok
    )


class _GLTF:
    ACCESSOR_TYPE_SIZE = {
        "SCALAR": 1,
        "VEC2": 2,
        "VEC3": 3,
        "VEC4": 4,
        "MAT2": 4,
        "MAT3": 9,
        "MAT4": 16,
    }

    COMPONENT_TYPE = {
        5120: np.int8,
        5121: np.uint8,
        5122: np.int16,
        5123: np.uint16,
        5125: np.uint32,
        5126: np.float32,
    }

    ATTRIBUTE_NAME = {
        "POSITION": "positions",
        "NORMAL": "normals",
        "TANGENT": "tangents",
        "TEXCOORD_0": "texcoords",
        "TEXCOORD_1": "texcoords1",
        "COLOR_0": "colors",
        "JOINTS_0": "skin_indices",
        "WEIGHTS_0": "skin_weights",
    }

    WRAP_MODE = {
        33071: "clamp-to-edge",  # CLAMP_TO_EDGE
        33648: "mirror-repeat",  # MIRRORED_REPEAT
        10497: "repeat",  # REPEAT
    }

    def __init__(self):
        self.scene = None
        self.scenes = []
        self.cameras = []
        self.lights = []
        self.animations = None

        self._plugins = {}

        self._register_plugin(GLTFMeshQuantizationExtension)
        self._register_plugin(GLTFMaterialsIorExtension)
        self._register_plugin(GLTFMaterialsSpecularExtension)
        self._register_plugin(GLTFMaterialsClearcoatExtension)
        self._register_plugin(GLTFMaterialsIridescenceExtension)
        self._register_plugin(GLTFMaterialsSheenExtension)
        self._register_plugin(GLTFMaterialsAnisotropyExtension)
        self._register_plugin(GLTFMaterialsEmissiveStrengthExtension)
        self._register_plugin(GLTFMaterialsUnlitExtension)
        self._register_plugin(GLTFLightsExtension)
        self._register_plugin(GLTFTextureTransformExtension)
        self._register_plugin(GLTFTextureWebPExtension)
        self._register_plugin(GLTFDracoMeshCompressionExtension)

    def _register_plugin(self, plugin_class):
        plugin = plugin_class(self)
        self._plugins[plugin.name] = plugin

    def load(self, path, quiet=False, remote_ok=True):
        """Load the whole gltf file, including meshes, skeletons, cameras, and animations."""
        self.__inner_load(path, quiet, remote_ok)

        self.scenes = self._load_scenes()
        if self._gltf.model.scene is not None:
            self.scene = self.scenes[self._gltf.model.scene]
        if self._gltf.model.animations is not None:
            self.animations = self._load_animations()

        return self

    async def load_async(self, path, quiet=False, remote_ok=True):
        """Load the whole gltf file, including meshes, skeletons, cameras, and animations."""
        await self.__inner_load_async(path, quiet, remote_ok)

        self.scenes = self._load_scenes()
        if self._gltf.model.scene is not None:
            self.scene = self.scenes[self._gltf.model.scene]
        if self._gltf.model.animations is not None:
            self.animations = self._load_animations()

        return self

    def load_mesh(self, path, quiet=False, materials=True, remote_ok=True):
        """Only load meshes from a gltf file, without skeletons, and no transformations applied."""

        self.__inner_load(path, quiet, remote_ok)

        meshes = []
        for gltf_mesh in self._gltf.model.meshes:
            mesh = self._load_gltf_mesh_by_info(gltf_mesh, load_material=materials)
            meshes.extend(mesh)
        return meshes

    async def load_mesh_async(self, path, quiet=False, materials=True, remote_ok=True):
        """Only load meshes from a gltf file, without skeletons, and no transformations applied."""

        await self.__inner_load_async(path, quiet, remote_ok)

        meshes = []
        for gltf_mesh in self._gltf.model.meshes:
            mesh = self._load_gltf_mesh_by_info(gltf_mesh, load_material=materials)
            meshes.extend(mesh)
        return meshes

    def __inner_load(self, path, quiet=False, remote_ok=True):
        if not find_spec("gltflib"):
            raise ImportError(
                "The `gltflib` library is required to load gltf scene: pip install gltflib"
            )
        import gltflib

        if "https://" in str(path) or "http://" in str(path):
            if not remote_ok:
                raise ValueError(
                    "Loading meshes from URLs is disabled. "
                    "Set remote_ok=True to allow loading from URLs."
                )
            if not find_spec("httpx"):
                raise ImportError(
                    "The `httpx` library is required to load meshes from URLs: pip install httpx"
                )

            import httpx
            from io import BytesIO
            from os import path as os_path
            import urllib.parse
            import mimetypes

            # download
            response = httpx.get(path, follow_redirects=True)
            response.raise_for_status()

            file_obj = BytesIO(response.content)

            ext = os_path.splitext(path)[1].lower()
            if ext == ".gltf":
                self._gltf = gltflib.GLTF.read_gltf(file_obj, load_file_resources=False)

            elif ext == ".glb":
                self._gltf = gltflib.GLTF.read_glb(file_obj, load_file_resources=False)

            downloadable_resources = [
                res
                for res in self._gltf.resources
                if isinstance(res, gltflib.FileResource)
            ]

            if downloadable_resources:
                for res in downloadable_resources:
                    res_path = urllib.parse.urljoin(path, res.uri)
                    try:
                        response = httpx.get(res_path)
                        response.raise_for_status()
                        res_file = BytesIO(response.content)

                        res._data = res_file.read()
                        res._mimetype = (
                            res._mimetype or mimetypes.guess_type(res.uri)[0]
                        )
                        res._loaded = True
                    except httpx.HTTPStatusError as e:
                        gfx.utils.logger.warning(f"download failed: {e} - {res_path}")
                    except Exception as e:
                        gfx.utils.logger.warning(f"download failed: {e} - {res_path}")

        else:  # local file
            self._gltf = gltflib.GLTF.load(path, load_file_resources=True)

        self.__post_inner_load(quiet)

    async def __inner_load_async(self, path, quiet=False, remote_ok=True):
        if not find_spec("gltflib"):
            raise ImportError(
                "The `gltflib` library is required to load gltf scene: pip install gltflib"
            )
        import gltflib

        if "https://" in str(path) or "http://" in str(path):
            if not remote_ok:
                raise ValueError(
                    "Loading meshes from URLs is disabled. "
                    "Set remote_ok=True to allow loading from URLs."
                )
            if not find_spec("httpx"):
                raise ImportError(
                    "The `httpx` library is required to load meshes from URLs: pip install httpx"
                )

            import httpx
            from io import BytesIO
            from os import path as os_path
            import urllib.parse
            import mimetypes
            import asyncio

            async with httpx.AsyncClient() as client:
                response = await client.get(path, follow_redirects=True)
                response.raise_for_status()

                file_obj = BytesIO(response.content)

                ext = os_path.splitext(path)[1].lower()
                if ext == ".gltf":
                    self._gltf = gltflib.GLTF.read_gltf(
                        file_obj, load_file_resources=False
                    )

                elif ext == ".glb":
                    self._gltf = gltflib.GLTF.read_glb(
                        file_obj, load_file_resources=False
                    )

                downloadable_resources = [
                    res
                    for res in self._gltf.resources
                    if isinstance(res, gltflib.FileResource)
                ]

                if downloadable_resources:

                    async def download_resource(res, client: httpx.AsyncClient):
                        res_path = urllib.parse.urljoin(path, res.uri)
                        try:
                            response = await client.get(res_path)
                            response.raise_for_status()
                            res_file = BytesIO(response.content)

                            res._data = res_file.read()
                            res._mimetype = (
                                res._mimetype or mimetypes.guess_type(res.uri)[0]
                            )
                            res._loaded = True

                            return res
                        except httpx.HTTPStatusError as e:
                            gfx.utils.logger.warning(
                                f"download failed: {e} - {res_path}"
                            )
                        except Exception as e:
                            gfx.utils.logger.warning(
                                f"download failed: {e} - {res_path}"
                            )

                    tasks = [
                        download_resource(res, client) for res in downloadable_resources
                    ]
                    await asyncio.gather(*tasks)

        else:  # local file
            self._gltf = gltflib.GLTF.load(path, load_file_resources=True)

        self.__post_inner_load(quiet)

    def __post_inner_load(self, quiet):
        if not quiet:
            extensions_required = self._gltf.model.extensionsRequired or []

            supported_extensions = self._plugins.keys()

            unsupported_extensions_required = set(extensions_required) - set(
                supported_extensions
            )

            if unsupported_extensions_required:
                gfx.utils.logger.warning(
                    f"This GLTF required extensions: {unsupported_extensions_required}, which are not supported yet."
                )
                # rasise or ignore?
                # raise NotImplementedError(f"This GLTF required extensions: {extensions_required}, which are not supported yet.")

            extensions_used = self._gltf.model.extensionsUsed or []

            unsupported_extensions_used = set(extensions_used) - set(
                supported_extensions
            )
            if unsupported_extensions_used:
                gfx.utils.logger.warning(
                    f"This GLTF used extensions: {unsupported_extensions_used}, which are not supported yet, so the display may not be so correct."
                )

        # mark the node types
        self._node_marks = self._mark_nodes()

    @lru_cache(maxsize=None)
    def _get_resource_by_uri(self, uri):
        for resource in self._gltf.resources:
            if resource.uri == uri:
                return resource
        raise ValueError(f"Buffer data not found for buffer {uri}")

    def _mark_nodes(self):
        gltf = self._gltf
        node_marks = [None] * len(gltf.model.nodes)

        # Nothing in the node definition indicates whether it is a Bone.
        # Use the skins' joint references to mark bones.
        if gltf.model.skins:
            for skin in gltf.model.skins:
                for joint in skin.joints:
                    node_marks[joint] = "Bone"

        # Meshes are marked when they are loaded
        # Maybe mark lights and other special nodes here
        return node_marks

    def _load_scenes(self):
        gltf = self._gltf
        scenes = []
        for scene in gltf.model.scenes:
            scene_obj = gfx.Group()
            scene_obj.name = scene.name
            scenes.append(scene_obj)

            for node in scene.nodes:
                node_obj = self._load_node(node)
                scene_obj.add(node_obj)

        return scenes

    @lru_cache(maxsize=None)
    def _load_node(self, node_index):
        gltf = self._gltf
        node_marks = self._node_marks

        node = gltf.model.nodes[node_index]

        translation = node.translation or [0, 0, 0]
        rotation = node.rotation or [0, 0, 0, 1]
        scale = node.scale or [1, 1, 1]

        if node.matrix is not None:
            matrix = np.array(node.matrix).reshape(4, 4).T
        else:
            matrix = la.mat_compose(translation, rotation, scale)

        node_mark = node_marks[node_index]

        if node_mark == "Bone":
            node_obj = gfx.Bone()

        elif node.camera is not None:
            camera_info = gltf.model.cameras[node.camera]
            if camera_info.type == "perspective":
                node_obj = gfx.PerspectiveCamera(
                    camera_info.perspective.yfov,
                    camera_info.perspective.aspectRatio,
                    depth_range=(
                        camera_info.perspective.znear,
                        camera_info.perspective.zfar,
                    ),
                )
            elif camera_info.type == "orthographic":
                node_obj = gfx.OrthographicCamera(
                    camera_info.orthographic.xmag,
                    camera_info.orthographic.ymag,
                    depth_range=(
                        camera_info.orthographic.znear,
                        camera_info.orthographic.zfar,
                    ),
                )
            else:
                raise ValueError(f"Unsupported camera type: {camera_info.type}")

            self.cameras.append(node_obj)
        elif node.mesh is not None:  # Mesh or SkinnedMesh
            # meshes = self._load_gltf_mesh(node.mesh, node.skin)
            # Do not use mesh cache here, we need to create a new mesh object for each node.
            mesh_info = self._gltf.model.meshes[node.mesh]
            meshes = self._load_gltf_mesh_by_info(mesh_info, node.skin)
            if node.weights:
                for gfx_mesh in meshes:
                    gfx_mesh.morph_target_influences = node.weights
            if len(meshes) == 1:
                node_obj = meshes[0]
            else:
                node_obj = gfx.Group()
                for mesh in meshes:
                    node_obj.add(mesh)
        else:
            node_obj = gfx.WorldObject()

        if node.extensions:
            for extension in node.extensions:
                if extension in self._plugins:
                    plugin = self._plugins[extension]
                    if hasattr(plugin, "create_node_attachment"):
                        node_attachment = plugin.create_node_attachment(node_index)
                        if node_attachment is not None:
                            node_obj.add(node_attachment)

        node_obj.local.matrix = matrix
        node_obj.name = node.name

        if node.children:
            for child in node.children:
                child_obj = self._load_node(child)
                node_obj.add(child_obj)

        return node_obj

    def _load_gltf_mesh_by_info(self, mesh, skin_index=None, load_material=True):
        meshes = []
        for primitive in mesh.primitives:
            geometry = self._load_gltf_geometry(primitive)
            primitive_mode = primitive.mode

            if load_material and primitive.material is not None:
                material = self._load_gltf_material(primitive.material)
            else:
                material = gfx.MeshStandardMaterial()
                if hasattr(geometry, "colors"):
                    material.color_mode = "vertex"

            if primitive_mode is None:
                primitive_mode = 4  # default to triangles

            if primitive_mode == 0:
                gfx_mesh = gfx.Points(geometry, material)

            elif primitive_mode in (1, 2, 3):
                # todo: distinguish lines, line_strip, line_loop
                gfx_mesh = gfx.Line(geometry, material)

            elif primitive_mode in (4, 5):
                # todo: distinguish triangles, triangle_strip, triangle_fan
                if (
                    skin_index is not None
                    and hasattr(geometry, "skin_indices")
                    and hasattr(geometry, "skin_weights")
                ):
                    # If the geometry has skin indices and weights, create a skinned mesh
                    gfx_mesh = gfx.SkinnedMesh(geometry, material)
                    skeleton = self._load_skins(skin_index)
                    gfx_mesh.bind(skeleton, np.identity(4))
                else:
                    gfx_mesh = gfx.Mesh(geometry, material)
            else:
                raise ValueError(f"Unsupported primitive mode: {primitive.mode}")

            if (
                getattr(geometry, "morph_positions", None)
                or getattr(geometry, "morph_normals", None)
                or getattr(geometry, "morph_colors", None)
            ):
                self.update_morph_target(gfx_mesh, mesh)

            meshes.append(gfx_mesh)

        return meshes

    def update_morph_target(self, gfx_mesh, mesh_def):
        if mesh_def.weights:
            gfx_mesh.morph_target_influences = mesh_def.weights

        if mesh_def.extras and mesh_def.extras.get("targetNames", None):
            gfx_mesh.morph_target_names.extend(mesh_def.extras["targetNames"])

    @lru_cache(maxsize=None)
    def _load_gltf_mesh(self, mesh_index, skin_index=None, load_material=True):
        mesh = self._gltf.model.meshes[mesh_index]
        return self._load_gltf_mesh_by_info(mesh, skin_index, load_material)

    @lru_cache(maxsize=None)
    def _load_gltf_material(self, material_index):
        material = self._gltf.model.materials[material_index]

        extensions = material.extensions or {}

        if (
            GLTFMaterialsUnlitExtension.EXTENSION_NAME in extensions
            and GLTFMaterialsUnlitExtension.EXTENSION_NAME in self._plugins
        ):
            plugin = self._plugins[GLTFMaterialsUnlitExtension.EXTENSION_NAME]
            material_type = plugin.get_material_type(material)
            gfx_material = material_type()
            plugin.extend_material(material, gfx_material)

        else:  # PBR
            material_type = gfx.MeshStandardMaterial

            # check if any plugin can handle this material type
            for extension in extensions:
                if extension in self._plugins:
                    plugin = self._plugins[extension]
                    if hasattr(plugin, "get_material_type"):
                        _material_type = plugin.get_material_type(material)
                        if _material_type is not None:
                            material_type = _material_type
                            break

            gfx_material = material_type()

            # check if any plugin can extend the material
            for extension in extensions:
                if extension in self._plugins:
                    plugin = self._plugins[extension]
                    if hasattr(plugin, "extend_material"):
                        plugin.extend_material(material, gfx_material)

            pbr_metallic_roughness = material.pbrMetallicRoughness
            if pbr_metallic_roughness is not None:
                if pbr_metallic_roughness.baseColorFactor is not None:
                    gfx_material.color = gfx.Color.from_physical(
                        *pbr_metallic_roughness.baseColorFactor
                    )

                if pbr_metallic_roughness.baseColorTexture is not None:
                    gfx_material.map = self._load_gltf_texture_map(
                        pbr_metallic_roughness.baseColorTexture
                    )

                if pbr_metallic_roughness.metallicRoughnessTexture is not None:
                    metallic_roughness_map = self._load_gltf_texture_map(
                        pbr_metallic_roughness.metallicRoughnessTexture
                    )
                    gfx_material.roughness_map = metallic_roughness_map
                    gfx_material.metalness_map = metallic_roughness_map

                if pbr_metallic_roughness.roughnessFactor is not None:
                    gfx_material.roughness = pbr_metallic_roughness.roughnessFactor
                else:
                    gfx_material.roughness = 1.0

                if pbr_metallic_roughness.metallicFactor is not None:
                    gfx_material.metalness = pbr_metallic_roughness.metallicFactor
                else:
                    gfx_material.metalness = 1.0

            if material.normalTexture is not None:
                gfx_material.normal_map = self._load_gltf_texture_map(
                    material.normalTexture
                )
                scale_factor = material.normalTexture.scale
                if scale_factor is None:
                    scale_factor = 1.0

                gfx_material.normal_scale = (scale_factor, scale_factor)

            if material.occlusionTexture is not None:
                gfx_material.ao_map = self._load_gltf_texture_map(
                    material.occlusionTexture
                )

            if material.emissiveFactor is not None:
                gfx_material.emissive = gfx.Color.from_physical(
                    *material.emissiveFactor
                )

            if material.emissiveTexture is not None:
                gfx_material.emissive_map = self._load_gltf_texture_map(
                    material.emissiveTexture
                )

        if material.alphaMode == "BLEND":
            gfx_material.alpha_mode = "blend"
            gfx_material.depth_write = False
        else:
            gfx_material.alpha_mode = "solid"
            if material.alphaMode == "MASK":
                if material.alphaCutoff is None:
                    gfx_material.alpha_test = 0.5
                else:
                    gfx_material.alpha_test = material.alphaCutoff

        gfx_material.side = (
            gfx.enums.VisibleSide.both
            if material.doubleSided
            else gfx.enums.VisibleSide.front
        )

        return gfx_material

    def _load_gltf_texture_map(self, texture_info):
        if isinstance(texture_info, dict):
            texture_index = texture_info["index"]
            uv_channel = texture_info.get("texCoord", 0)
            extensions = texture_info.get("extensions", {})
        else:
            texture_index = texture_info.index
            uv_channel = texture_info.texCoord or 0
            extensions = texture_info.extensions or {}

        texture = self._load_gltf_texture(texture_index)
        if texture is None:
            return None

        texture_map = gfx.TextureMap(texture, uv_channel=uv_channel)

        texture_desc = self._gltf.model.textures[texture_index]

        sampler = texture_desc.sampler

        if sampler is not None:
            sampler = self._load_gltf_sampler(sampler)

            if sampler.magFilter == 9728:
                texture_map.mag_filter = "nearest"
            elif sampler.magFilter == 9729 or sampler.magFilter is None:
                texture_map.mag_filter = "linear"

            if sampler.minFilter == 9728:  # NEAREST
                texture_map.min_filter = "nearest"
            elif sampler.minFilter == 9729:  # LINEAR
                texture_map.min_filter = "linear"
            elif sampler.minFilter == 9984:  # NEAREST_MIPMAP_NEAREST
                texture._generate_mipmaps = True
                texture_map.min_filter = "nearest"
                texture_map.mipmap_filter = "nearest"
            elif sampler.minFilter == 9985:  # LINEAR_MIPMAP_NEAREST
                texture._generate_mipmaps = True
                texture_map.min_filter = "linear"
                texture_map.mipmap_filter = "nearest"
            elif sampler.minFilter == 9986:  # NEAREST_MIPMAP_LINEAR
                texture._generate_mipmaps = True
                texture_map.min_filter = "nearest"
                texture_map.mipmap_filter = "linear"
            elif (
                sampler.minFilter == 9987 or sampler.minFilter is None
            ):  # LINEAR_MIPMAP_LINEAR
                texture._generate_mipmaps = True
                texture_map.min_filter = "linear"
                texture_map.mipmap_filter = "linear"

            texture_map.wrap_s = self.WRAP_MODE[sampler.wrapS or 10497]
            texture_map.wrap_t = self.WRAP_MODE[sampler.wrapT or 10497]

        for extension in extensions:
            if extension in self._plugins:
                plugin = self._plugins[extension]
                if hasattr(plugin, "extend_texture"):
                    plugin.extend_texture(texture_info, texture_map)

        return texture_map

    @lru_cache(maxsize=None)
    def _load_gltf_texture(self, texture_index):
        texture_desc = self._gltf.model.textures[texture_index]

        extensions = texture_desc.extensions or {}
        for extension in extensions:
            if extension in self._plugins:
                plugin = self._plugins[extension]
                if hasattr(plugin, "load_texture"):
                    return plugin.load_texture(texture_desc)

        source = texture_desc.source
        if source is None:
            return None
        image = self._load_image(source)
        texture = gfx.Texture(image, dim=2)
        return texture

    @lru_cache(maxsize=None)
    def _load_gltf_sampler(self, sampler_index):
        sampler = self._gltf.model.samplers[sampler_index]
        # Sampler( magFilter=9729, minFilter=9987, wrapS=None, wrapT=None)
        return sampler

    @lru_cache(maxsize=None)
    def _load_image(self, image_index):
        image_info = self._gltf.model.images[image_index]
        import imageio.v3 as iio

        if image_info.bufferView is not None:
            image_data = self._get_buffer_memory_view(image_info.bufferView)

        elif image_info.uri is not None:
            resource = self._get_resource_by_uri(image_info.uri)
            image_data = resource.data
        else:
            raise ValueError("No image data found")

        # need consider mimeType?
        image = iio.imread(image_data, mode="RGBA")
        return image

    def _load_gltf_geometry(self, primitive):
        extensions = primitive.extensions or {}

        geometry = None

        if extensions:
            for extension in extensions:
                if extension in self._plugins:
                    plugin = self._plugins[extension]
                    if hasattr(plugin, "load_geometry"):
                        geometry = plugin.load_geometry(primitive)

        if geometry is None:
            indices_accessor = primitive.indices
            geometry_args = {}

            for attr, accessor_index in primitive.attributes.__dict__.items():
                if accessor_index is not None:
                    if attr in self.ATTRIBUTE_NAME:
                        geometry_attr = self.ATTRIBUTE_NAME[attr]
                    else:
                        if attr.startswith("TEXCOORD_"):
                            geometry_attr = f"texcoords{attr[-1]}"
                        else:
                            geometry_attr = attr.lower()

                    data = self._load_accessor(accessor_index)

                    # pygfx not support int attributes now, so we need to convert them to float.
                    if geometry_attr in (
                        "positions",
                        "normals",
                        "tangents",
                        "texcoords",
                        "texcoords1",
                    ) or geometry_attr.startswith("texcoords"):
                        data = data.astype(np.float32, copy=False)

                    geometry_args[geometry_attr] = data

            if indices_accessor is not None:
                indices = self._load_accessor(indices_accessor).reshape(-1, 3)
            else:
                # TODO: For now, pygfx not support non-indexed geometry, so we need to generate indices for them.
                # Remove this after pygfx support non-indexed geometry.
                indices = np.arange(
                    len(geometry_args["positions"]) // 3 * 3, dtype=np.int32
                ).reshape((-1, 3))

            geometry_args["indices"] = indices

            geometry = gfx.Geometry(**geometry_args)

        if primitive.targets:
            for target in primitive.targets:
                # print(target)
                for attr, accessor_index in target.__dict__.items():
                    if accessor_index is not None:
                        target_attr = f"morph_{self.ATTRIBUTE_NAME[attr]}"
                        data = self._load_accessor(accessor_index).astype(
                            np.float32, copy=False
                        )

                        morph_attr = getattr(geometry, target_attr, None)
                        if morph_attr is None:
                            morph_attr = []
                            setattr(geometry, target_attr, morph_attr)

                        morph_attr.append(data)

            geometry.morph_targets_relative = True

        return geometry

    @lru_cache(maxsize=None)
    def _get_buffer_memory_view(self, buffer_view_index):
        gltf = self._gltf
        buffer_view = gltf.model.bufferViews[buffer_view_index]
        buffer = gltf.model.buffers[buffer_view.buffer]
        resource = self._get_resource_by_uri(buffer.uri)
        m = memoryview(resource.data)
        view = m[
            buffer_view.byteOffset : (buffer_view.byteOffset or 0)
            + buffer_view.byteLength
        ]
        return view

    @lru_cache(maxsize=None)
    def _load_accessor(self, accessor_index):
        gltf = self._gltf
        accessor = gltf.model.accessors[accessor_index]

        buffer_view = gltf.model.bufferViews[accessor.bufferView]
        view = self._get_buffer_memory_view(accessor.bufferView)

        # todo accessor.sparse
        if accessor.sparse is not None:
            gfx.utils.logger.warning("Sparse accessor is not supported yet.")

        accessor_type = accessor.type
        accessor_component_type = accessor.componentType
        accessor_count = accessor.count
        accessor_dtype = np.dtype(self.COMPONENT_TYPE[accessor_component_type])
        accessor_offset = accessor.byteOffset or 0
        accessor_type_size = self.ACCESSOR_TYPE_SIZE[accessor_type]

        item_bytes = accessor_type_size * accessor_dtype.itemsize

        if buffer_view.byteStride and buffer_view.byteStride != item_bytes:
            # It's a interleaved buffer
            # pygfx not support interleaved buffer now, so we pick out the data we need from the interleaved buffer.
            # TODO: optimize this after pygfx support interleaved buffer.
            ar = np.lib.stride_tricks.as_strided(
                view[accessor_offset:],
                shape=(accessor_count, item_bytes),
                strides=(buffer_view.byteStride, 1),
            )
            ar = np.frombuffer(np.ascontiguousarray(ar), dtype=accessor_dtype)
        else:
            ar = np.frombuffer(
                view,
                dtype=accessor_dtype,
                offset=accessor_offset,
                count=accessor_count * accessor_type_size,
            )
        if accessor_type_size > 1:
            ar = ar.reshape(accessor_count, accessor_type_size)

        if accessor.normalized:
            # KHR_mesh_quantization
            # https://github.com/KhronosGroup/glTF/tree/main/extensions/2.0/Khronos/KHR_mesh_quantization
            assert accessor_dtype.kind == "i" or accessor_dtype.kind == "u"
            ar = ar.astype(np.float32, copy=False) / np.iinfo(accessor_dtype).max

        # pygfx not support int8, int16, uint8, uint16 now
        if ar.dtype == np.uint8 or ar.dtype == np.uint16:
            ar = ar.astype(np.uint32, copy=False)
        if ar.dtype == np.int8 or ar.dtype == np.int16:
            ar = ar.astype(np.int32, copy=False)

        return ar

    @lru_cache(maxsize=None)
    def _load_skins(self, skin_index):
        skin = self._gltf.model.skins[skin_index]
        bones = [self._load_node(index) for index in skin.joints]
        inverse_bind_matrices = self._load_accessor(skin.inverseBindMatrices)

        bone_inverses = []
        for matrices in inverse_bind_matrices:
            bone_inverse = np.array(matrices).reshape(4, 4).T
            bone_inverses.append(bone_inverse)

        skeleton = gfx.Skeleton(bones, bone_inverses)
        return skeleton

    def _load_animation(self, animation_info):
        channels = animation_info.channels
        samplers = animation_info.samplers

        duration = 0

        key_frame_tracks = []

        for channel in channels:
            target = channel.target
            sampler = samplers[channel.sampler]

            if target.node is None:
                # todo: now we only support node animation
                gfx.utils.logger.warning(
                    f"Animation channel target [{target}] node is None, skip it for now."
                )
                continue

            target_node = self._load_node(target.node)
            name = target_node.name
            target_property = target.path
            interpolation = sampler.interpolation or "LINEAR"
            times = self._load_accessor(sampler.input)
            if times[-1] > duration:
                duration = times[-1]
            values = self._load_accessor(sampler.output)

            if interpolation == "LINEAR":
                if target_property == "rotation":
                    interpolation_fn = gfx.QuaternionLinearInterpolant
                else:
                    interpolation_fn = gfx.LinearInterpolant
            elif interpolation == "STEP":
                interpolation_fn = gfx.StepInterpolant
            elif interpolation == "CUBICSPLINE":
                # interpolation_fn = gfx.CubicSplineInterpolant
                # A CUBICSPLINE keyframe in glTF has three output values for each input value,
                # representing inTangent, splineVertex, and outTangent.
                interpolation_fn = (
                    GLTFCubicSplineInterpolant
                    if target_property != "rotation"
                    else GLTFCubicSplineQuaternionInterpolant
                )
            else:
                raise ValueError(f"Unsupported interpolation type: {interpolation}")

            if interpolation == "CUBICSPLINE":
                # Layout of keyframe output values for CUBICSPLINE animations:
                # [ inTangent_1, splineVertex_1, outTangent_1, inTangent_2, splineVertex_2, ... ]
                values = values.reshape(len(times), -1, values.shape[-1])
            else:
                values = values.reshape(len(times), -1)

            if len(times) != len(values):
                gfx.utils.logger.warning(
                    f"keyframe: {name}, times and values have different lengths, {len(times)} != {len(values)}"
                )
                length = min(len(times), len(values))
                times = times[:length]
                values = values[:length]

            keyframe = gfx.KeyframeTrack(
                name, target_node, target_property, times, values, interpolation_fn
            )
            key_frame_tracks.append(keyframe)

        action_clip = gfx.AnimationClip(animation_info.name, duration, key_frame_tracks)

        return action_clip

    def _load_animations(self):
        gltf = self._gltf
        animations = []
        for animation in gltf.model.animations:
            action_clip = self._load_animation(animation)
            animations.append(action_clip)
        return animations


def print_scene_graph(obj, show_pos=False, show_rot=False, show_scale=False):
    """
    Print the tree structure of the scene, including the optional position, rotation, and scale of each object.

    Parameters:
    ----------
    obj : gfx.WorldObject
        The root object.
    show_pos : bool
        Whether to show the position of each object.
    show_rot : bool
        Whether to show the rotation of each object.
    show_scale : bool
        Whether to show the scale of each object.
    """

    def _print_tree(obj: gfx.WorldObject, level=0):
        name = "- " * level + f"{obj.__class__.__name__}[{obj.name}]"
        if show_pos:
            name += f"\n{'  ' * level}|- pos: {obj.local.position}"
        if show_rot:
            name += f"\n{'  ' * level}|- rot: {obj.local.rotation}"
        if show_scale:
            name += f"\n{'  ' * level}|- scale: {obj.local.scale}"

        print(name)

        for child in obj.children:
            _print_tree(child, level=level + 1)

    _print_tree(obj)


class GLTFCubicSplineInterpolant(gfx.Interpolant):
    """
    See: https://registry.khronos.org/glTF/specs/2.0/glTF-2.0.html#interpolation-cubic
    """

    def __init__(self, times, values):
        super().__init__(times, values)

    def _interpolate(self, i1, t0, t, t1):
        dt = t1 - t0

        p = (t - t0) / dt
        pp = p * p
        ppp = pp * p

        s2 = -2 * ppp + 3 * pp
        s3 = ppp - pp
        s0 = 1 - s2
        s1 = s3 - pp + p

        # Layout of keyframe output values for CUBICSPLINE animations:
        # [ [inTangent_1, splineVertex_1, outTangent_1], [inTangent_2, splineVertex_2, outTangent_2], ... ]

        values = self.sample_values
        p0 = values[i1 - 1][1]
        m0 = values[i1 - 1][2] * dt

        p1 = values[i1][1]
        m1 = values[i1][0] * dt

        return s0 * p0 + s1 * m0 + s2 * p1 + s3 * m1


class GLTFCubicSplineQuaternionInterpolant(GLTFCubicSplineInterpolant):
    def _interpolate(self, i1, t0, t, t1):
        res = super()._interpolate(i1, t0, t, t1)
        # remember normalize the quaternion
        return res / np.linalg.norm(res)


### GLTF Extensions ###


class GLTFExtension:
    """
    Abstract Base class for GLTF extensions.
    """

    EXTENSION_NAME = ""

    def __init__(self, parser: _GLTF):
        self.name = self.EXTENSION_NAME
        self.parser = parser


class GLTFMeshQuantizationExtension(GLTFExtension):
    EXTENSION_NAME = "KHR_mesh_quantization"


class GLTFBaseMaterialsExtension(GLTFExtension):
    """
    Abstract Base class for GLTF materials extensions.
    """

    MATERIAL_TYPE = gfx.MeshPhysicalMaterial

    def get_material_type(self, material_def):
        if material_def.extensions and self.EXTENSION_NAME in material_def.extensions:
            return self.MATERIAL_TYPE
        else:
            return None

    def extend_material(self, material_def, material):
        pass


class GLTFMaterialsIorExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_ior"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        material.ior = extension.get("ior", 1.5)


class GLTFMaterialsSpecularExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_specular"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        material.specular_intensity = extension.get("specularFactor", 1.0)

        specular_texture = extension.get("specularTexture", None)

        if specular_texture is not None:
            material.specular_intensity_map = self.parser._load_gltf_texture_map(
                specular_texture
            )

        specular_color = extension.get("specularColorFactor", [1.0, 1.0, 1.0])
        material.specular = gfx.Color.from_physical(*specular_color)

        specular_color_texture = extension.get("specularColorTexture", None)

        if specular_color_texture is not None:
            material.specular_map = self.parser._load_gltf_texture_map(
                specular_color_texture
            )


class GLTFMaterialsClearcoatExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_clearcoat"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        clearcoat_factor = extension.get("clearcoatFactor", None)
        if clearcoat_factor is not None:
            material.clearcoat = clearcoat_factor

        clearcoat_texture = extension.get("clearcoatTexture", None)
        if clearcoat_texture is not None:
            material.clearcoat_map = self.parser._load_gltf_texture_map(
                clearcoat_texture
            )

        clearcoat_roughness_factor = extension.get("clearcoatRoughnessFactor", None)
        if clearcoat_roughness_factor is not None:
            material.clearcoat_roughness = clearcoat_roughness_factor

        clearcoat_rughness_texture = extension.get("clearcoatRoughnessTexture", None)
        if clearcoat_rughness_texture is not None:
            material.clearcoat_roughness_map = self.parser._load_gltf_texture_map(
                clearcoat_rughness_texture
            )

        clearcoat_normal_texture = extension.get("clearcoatNormalTexture", None)
        if clearcoat_normal_texture is not None:
            material.clearcoat_normal_map = self.parser._load_gltf_texture_map(
                clearcoat_normal_texture
            )
            clearcoat_normal_scale = clearcoat_normal_texture.get("scale", None)
            if clearcoat_normal_scale is not None:
                material.clearcoat_normal_scale = (
                    clearcoat_normal_scale,
                    clearcoat_normal_scale,
                )


class GLTFMaterialsIridescenceExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_iridescence"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        iridescence_factor = extension.get("iridescenceFactor", None)
        if iridescence_factor is not None:
            material.iridescence = iridescence_factor

        iridescence_texture = extension.get("iridescenceTexture", None)
        if iridescence_texture is not None:
            material.iridescence_map = self.parser._load_gltf_texture_map(
                iridescence_texture
            )

        iridescence_ior = extension.get("iridescenceIor", None)
        material.iridescence_ior = iridescence_ior or 1.3

        iridescence_thickness_min = extension.get("iridescenceThicknessMinimum", 100)
        iridescence_thickness_max = extension.get("iridescenceThicknessMaximum", 400)

        material.iridescence_thickness_range = (
            iridescence_thickness_min,
            iridescence_thickness_max,
        )

        iridescence_thickness_texture = extension.get(
            "iridescenceThicknessTexture", None
        )
        if iridescence_thickness_texture is not None:
            material.iridescence_thickness_map = self.parser._load_gltf_texture_map(
                iridescence_thickness_texture
            )


class GLTFMaterialsAnisotropyExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_anisotropy"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        anisotropy_factor = extension.get("anisotropyStrength", None)
        if anisotropy_factor is not None:
            material.anisotropy = anisotropy_factor

        anisotropy_rotation = extension.get("anisotropyRotation", None)
        if anisotropy_rotation is not None:
            material.anisotropy_rotation = anisotropy_rotation

        anisotropy_texture = extension.get("anisotropyTexture", None)
        if anisotropy_texture is not None:
            material.anisotropy_map = self.parser._load_gltf_texture_map(
                anisotropy_texture
            )


class GLTFMaterialsSheenExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_sheen"

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]

        sheen_color = extension.get("sheenColorFactor", None)
        if sheen_color is not None:
            material.sheen = 1.0
            material.sheen_color = gfx.Color.from_physical(*sheen_color)

        sheen_color_texture = extension.get("sheenColorTexture", None)
        if sheen_color_texture is not None:
            material.sheen_color_map = self.parser._load_gltf_texture_map(
                sheen_color_texture
            )

        sheen_roughness_factor = extension.get("sheenRoughnessFactor", None)
        if sheen_roughness_factor is not None:
            material.sheen_roughness = sheen_roughness_factor

        sheen_roughness_texture = extension.get("sheenRoughnessTexture", None)
        if sheen_roughness_texture is not None:
            material.sheen_roughness_map = self.parser._load_gltf_texture_map(
                sheen_roughness_texture
            )


class GLTFMaterialsEmissiveStrengthExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_emissive_strength"
    MATERIAL_TYPE = None

    def get_material_type(self, material_def):
        return None

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        extension = material_def.extensions[self.EXTENSION_NAME]
        emissive_strength = extension.get("emissiveStrength", None)
        if emissive_strength is not None:
            material.emissive_intensity = emissive_strength


class GLTFMaterialsUnlitExtension(GLTFBaseMaterialsExtension):
    EXTENSION_NAME = "KHR_materials_unlit"
    MATERIAL_TYPE = gfx.MeshBasicMaterial

    def extend_material(self, material_def, material):
        if (
            not material_def.extensions
            or self.EXTENSION_NAME not in material_def.extensions
        ):
            return

        pbr_metallic_roughness = material_def.pbrMetallicRoughness
        if pbr_metallic_roughness is not None:
            if pbr_metallic_roughness.baseColorFactor is not None:
                material.color = gfx.Color.from_physical(
                    *pbr_metallic_roughness.baseColorFactor
                )

            if pbr_metallic_roughness.baseColorTexture is not None:
                material.map = self.parser._load_gltf_texture_map(
                    pbr_metallic_roughness.baseColorTexture
                )


class GLTFLightsExtension(GLTFExtension):
    EXTENSION_NAME = "KHR_lights_punctual"

    def create_node_attachment(self, node_index):
        """
        Create a light object and attach it to the node.
        """
        gltf = self.parser._gltf
        node = gltf.model.nodes[node_index]
        if node.extensions and self.EXTENSION_NAME in node.extensions:
            extension = node.extensions[self.EXTENSION_NAME]
            light_index = extension["light"]

            light = self._load_light(light_index)

            if light is not None:
                self.parser.lights.append(light)

            return light

    @lru_cache(maxsize=None)
    def _load_light(self, light_index):
        gltf = self.parser._gltf

        extensions = (gltf.model.extensions and gltf.model.extensions[self.name]) or {}
        light_defs = extensions.get("lights", None)

        if light_defs:
            if light_index >= len(light_defs):
                return None

            light_info = light_defs[light_index]

            color = light_info.get("color", [1, 1, 1])
            light_color = gfx.Color.from_physical(*color)
            light_range = light_info.get("range", 0)
            intensity = light_info.get("intensity", 1.0)

            light_type = light_info["type"]

            if light_type == "directional":
                light = gfx.DirectionalLight(color=light_color, intensity=intensity)
                light.target.local.position = (0, 0, -1)
                light.add(light.target)
            elif light_type == "point":
                light = gfx.PointLight(color=light_color, intensity=intensity)
                light.distance = light_range
            elif light_type == "spot":
                light = gfx.SpotLight(color=light_color, intensity=intensity)
                light.distance = light_range

                if "spot" in light_info:
                    spot = light_info["spot"]
                    inner_cone_angle = spot.get("innerConeAngle", 0)
                    outer_cone_angle = spot.get("innerConeAngle", np.pi / 4)
                    light.angle = outer_cone_angle
                    light.penumbra = 1.0 - (inner_cone_angle / outer_cone_angle)
                    light.target.local.position = (0, 0, -1)
                    light.add(light.target)

            else:
                raise ValueError(f"Unsupported light type: {light_info.type}")

            light.local.position = (0, 0, 0)
            light.name = light_info.get("name", f"light_{light_index}")

            return light


class GLTFTextureTransformExtension(GLTFExtension):
    EXTENSION_NAME = "KHR_texture_transform"

    def extend_texture(self, texture_info, texture_map):
        if isinstance(texture_info, dict):
            extensions = texture_info.get("extensions", None)
        else:
            extensions = texture_info.extensions

        if self.EXTENSION_NAME not in extensions:
            return

        extension = extensions[self.EXTENSION_NAME]

        texture_map.update_matrix_components(
            offset=extension.get("offset", None),
            rotation=extension.get("rotation", None),
            scale=extension.get("scale", None),
        )

        texcoord = extension.get("texCoord", None)
        if texcoord is not None:
            texture_map.uv_channel = texcoord


class GLTFTextureWebPExtension(GLTFExtension):
    EXTENSION_NAME = "EXT_texture_webp"

    def load_texture(self, texture_desc):
        extensions = texture_desc.extensions or {}
        if self.EXTENSION_NAME not in extensions:
            return

        extension = extensions[self.EXTENSION_NAME]

        source = extension.get("source", None)
        if source is None:
            source = texture_desc.source
        image = self.parser._load_image(source)
        texture = gfx.Texture(image, dim=2)
        return texture


class GLTFDracoMeshCompressionExtension(GLTFExtension):
    EXTENSION_NAME = "KHR_draco_mesh_compression"

    def __init__(self, parser: _GLTF):
        super().__init__(parser)

    def load_geometry(self, primitive):
        if (
            primitive.extensions is None
            or self.EXTENSION_NAME not in primitive.extensions
        ):
            return None

        if not find_spec("DracoPy"):
            raise ImportError(
                """The `DracoPy` library is required for loading Draco compressed meshes. \n
                Please install it with `pip install DracoPy`."""
            )

        import DracoPy

        draco_extension = primitive.extensions[self.EXTENSION_NAME]
        buffer_view_index = draco_extension["bufferView"]

        buffer_view = self.parser._get_buffer_memory_view(buffer_view_index)
        draco_mesh = DracoPy.decode(bytes(buffer_view))

        geometry_args = {}

        if draco_mesh.points is not None:
            geometry_args["positions"] = draco_mesh.points.astype(
                np.float32, copy=False
            )
        if draco_mesh.normals is not None:
            geometry_args["normals"] = draco_mesh.normals.astype(np.float32, copy=False)
        if draco_mesh.colors is not None:
            geometry_args["colors"] = draco_mesh.colors.astype(np.float32, copy=False)
        if draco_mesh.tex_coord is not None:
            geometry_args["texcoords"] = draco_mesh.tex_coord.astype(
                np.float32, copy=False
            )

        if draco_mesh.faces is not None:
            geometry_args["indices"] = draco_mesh.faces.astype(np.uint32, copy=False)
        else:
            geometry_args["indices"] = np.arange(
                len(geometry_args["positions"]) // 3 * 3, dtype=np.int32
            ).reshape((-1, 3))

        geometry = gfx.Geometry(**geometry_args)
        return geometry



"""
A global object shared by all renderers.
"""

import os
import wgpu

from ....resources import Resource, Buffer
from ....utils.trackable import Trackable
from ....utils import array_from_shadertype
from ....utils.text import glyph_atlas

from .utils import gpu_caches


# Definition uniform struct with standard info related to transforms,
# provided to each shader as uniform at slot 0.
# todo: a combined transform would be nice too, for performance
# todo: same for ndc_to_world transform (combined inv transforms)
stdinfo_uniform_type = dict(
    cam_transform="4x4xf4",
    cam_transform_inv="4x4xf4",
    projection_transform="4x4xf4",
    projection_transform_inv="4x4xf4",
    ndc_offset="4xf4",
    physical_size="2xf4",
    logical_size="2xf4",
)


class Shared(Trackable):
    """An object to store global data to share between multiple wgpu
    renderers. Each renderer updates the data and then passes this down
    to the pipeline containers.

    The renderer instantiates and stores the singleton shared object.
    """

    # Vanilla WGPU does not support interpolating samplers for float32
    # textures, which is sad for e.g. volume rendering. We require the
    # 'float32-filterable' feature for this.
    #
    # Previously (when this feature was not yet available) we used the
    # native-only feature 'texture_adapter_specific_format_features',
    # but that one enables far more than we want.
    #
    # Technically we can check if the feature is available and fall
    # back (in some way) when it's not available, but this easily
    # results in complex code. I think we can probably get away with
    # requiring only a few features that are available on the main
    # target platforms.

    _features = set(["float32-filterable"])
    _limits = dict()
    _selected_adapter = None
    _power_preference = None
    _instance = None

    def __init__(self, *, canvas=None):
        super().__init__()

        # Set this instance as the global one
        assert Shared._instance is None
        Shared._instance = self

        # Select adapter to use.
        if Shared._selected_adapter:
            self._adapter = Shared._selected_adapter
        elif adapter_name := os.environ.get("PYGFX_WGPU_ADAPTER_NAME"):
            # Similar to https://github.com/gfx-rs/wgpu?tab=readme-ov-file#environment-variables
            adapters = wgpu.gpu.enumerate_adapters_sync()
            adapters_llvm = [a for a in adapters if adapter_name in a.summary]
            if not adapters_llvm:
                raise ValueError(f"Adapter with name '{adapter_name}' not found.")
            self._adapter = adapters_llvm[0]
        else:
            self._adapter = wgpu.gpu.request_adapter_sync(
                power_preference=Shared._power_preference or "high-performance"
            )

        # Create logical device from adapter. There should be just one per
        # process. Having a global device provides the benefit that we can draw
        # any object anywhere. Supporting different devices per renderer/canvas
        # is technically possible, but would require an extra layer of
        # indirection in the pipeline objects (device -> renderstate -> passes).
        # So out of scope for the time being.
        self._device = self.adapter.request_device_sync(
            required_features=list(Shared._features), required_limits=Shared._limits
        )

        self._create_diagnostics()

        # Create a uniform buffer for std info
        # Stored on _store so if we'd ever swap it out for another buffer,
        # the pipeline automatically update.
        self._store.uniform_buffer = Buffer(
            array_from_shadertype(stdinfo_uniform_type), force_contiguous=True
        )
        self._store.uniform_buffer._wgpu_usage |= wgpu.BufferUsage.UNIFORM

        # Init glyph atlas texture
        self._store.glyph_atlas_texture = None
        self._store.glyph_atlas_info_buffer = None
        self.pre_render_hook()

    def pre_render_hook(self):
        """Called by the renderer on the beginning of a draw."""
        tex = glyph_atlas.texture
        if tex is not self._store["glyph_atlas_texture"]:
            self._store.glyph_atlas_texture = tex
        buffer = glyph_atlas.info_buffer
        if buffer is not self._store["glyph_atlas_info_buffer"]:
            self._store.glyph_atlas_info_buffer = buffer

    @classmethod
    def get_instance(cls):
        return cls._instance

    @property
    def adapter(self):
        """The shared WGPU adapter object."""
        return self._adapter

    @property
    def device(self):
        """The shared WGPU device object."""
        return self._device

    @property
    def uniform_buffer(self):
        """The shared uniform buffer in which the renderer puts
        information about the canvas and camera.
        """
        return self._store.uniform_buffer

    @property
    def glyph_atlas_texture(self):
        """The shared glyph atlas (as a texture view and sampler) for objects that want to render text."""
        return self._store.glyph_atlas_texture

    @property
    def glyph_atlas_info_buffer(self):
        """A buffer containing per-glyph metadata (rects and more)."""
        return self._store.glyph_atlas_info_buffer

    def _create_diagnostics(self):
        # Since diagnostics objects, are shown in order of creation,
        # its good to instantiate our diagnostics objects right after loading
        # the wgpu_native backend from wgpu.
        PyGfxAdapterInfoDiagnostics("pygfx_adapter_info")
        PyGfxFeaturesDiagnostics("pygfx_features")
        PyGfxLimitsDiagnostics("pygfx_limits")
        PygfxCacheDiagnostics("pygfx_caches")
        PygfxResourceDiagnostics("pygfx_resources")


def select_power_preference(power_preference):
    """Select whether a powerful or battery-friendly GPU is selected.

    Accepts a value from ``wgpu.PowerPreference``: "high-performance" or "low-power".

    This function must be called before before the first ``Renderer`` is created.
    """
    if power_preference not in wgpu.PowerPreference:
        raise ValueError(
            f"select_power_preference() received invalid value for {wgpu.PowerPreference!r}."
        )
    if Shared._instance is not None:
        raise RuntimeError(
            "The select_power_preference() function must be called before creating the first renderer."
        )
    Shared._power_preference = power_preference


def select_adapter(adapter):
    """Select a specific adapter / GPU.

    Select an adapter as obtained via ``wgpu.gpu.enumerate_adapters_sync()``, which
    can be useful in multi-gpu environments.

    For example::

        adapters = wgpu.gpu.enumerate_adapters_sync()
        adapters_tesla = [a for a in adapters if "Tesla" in a.summary]
        adapters_discrete = [a for a in adapters if "DiscreteGPU" in a.summary]
        pygfx.renderers.wgpu.select_adapter(adapters_discrete[0])

    Note that using this function reduces the portability of your code, because
    it's highly specific for your current machine/environment.

    The order of the adapters returned by ``wgpu.gpu.enumerate_adapters_sync()`` is
    such that Vulkan adapters go first, then Metal, then D3D12, then OpenGL.
    Within each category, the order as provided by the particular backend is
    maintained. Note that the same device may be present via multiple backends
    (e.g. vulkan/opengl).

    We cannot make guarantees about whether the order of the adapters matches
    the order as reported by e.g. ``nvidia-smi``. We have found that on a Linux
    multi-gpu cluster, the order does match, but we cannot promise that this is
    always the case. If you want to make sure, do some testing by allocating big
    buffers and checking memory usage using ``nvidia-smi``.

    Example to allocate and check GPU mem usage::

        import subprocess

        import wgpu
        import torch

        def allocate_gpu_mem_with_wgpu(idx):
            a = wgpu.gpu.enumerate_adapters_sync()[idx]
            d = a.request_device_sync()
            b = d.create_buffer(size=10*2**20, usage=wgpu.BufferUsage.COPY_DST)
            return b

        def allocate_gpu_mem_with_torch(idx):
            d = torch.device(f"cuda:{idx}")
            return torch.ones([2000, 10], dtype=torch.float32, device=d)

        def show_mem_usage():
            print(subprocess.run(["nvidia-smi"]))

    See https://github.com/pygfx/wgpu-py/issues/482 for more details.

    """
    if not isinstance(adapter, wgpu.GPUAdapter):
        raise TypeError(
            f"select_adapter() only accepts a wgpu.GPUAdapter object, but got {adapter.__class__.__name__}."
        )
    if Shared._instance is not None:
        raise RuntimeError(
            "The select_adapter() function must be called before creating the first renderer."
        )
    Shared._selected_adapter = adapter


def enable_wgpu_features(*features):
    """Enable specific features (as strings) on the wgpu device.

    WARNING: enabling features means that your code may not work on all
    devices. The point of wgpu is that it can make a promise that a
    visualization works and looks the same on any device. Using features
    breaks that promise, and may cause your code to not work on e.g.
    mobile devices or certain operating systems.

    Features can also be turned off by prefixing with "!". This e.g. allows
    using a subset of Pygfx on old hardware by using
    ``enable_wgpu_features("!float32-filterable")``. Use with care; some parts
    of Pygfx will not function without the default features enabled.

    This function must be called before before the first ``Renderer`` is created.
    It can be called multiple times to enable more features. Note that feature names
    are invariant to use of dashes versus underscores.

    For more information on features:

    * ``wgpu.FeatureName`` for all possible official features.
    * ``renderer.device.adapter.features`` for the features available on the current system.
    * ``renderer.device.features`` for the currently enabled features.
    * https://gpuweb.github.io/gpuweb/#gpufeaturename for the official webgpu features (excl. native features).
    * https://docs.rs/wgpu/latest/wgpu/struct.Features.html for the features and their limitations in wgpu-core.
    """
    if Shared._instance is not None:
        raise RuntimeError(
            "The enable_wgpu_features() function must be called before creating the first renderer."
        )
    # Split into features to turn on or off
    features = set(features)
    features_off = {f for f in features if f.startswith("!")}
    features_on = features - features_off
    # Make the off-features work on different spellings.
    # As it is now, this does not discard the "float32Filterable" spelling.
    features_off = {f.lstrip("!") for f in features_off}
    features_off |= {f.replace("-", "_") for f in features_off}
    features_off |= {f.replace("_", "-") for f in features_off}
    # Apply
    Shared._features.difference_update(features_off)
    Shared._features.update(features_on)


def set_wgpu_limits(**limits):
    """Set specific limits (as key-value pairs) on the wgpu device.

    WARNING: setting high limits may make your code less portable across devices.

    This function must be called before before the first ``Renderer`` is created.
    It can be called multiple times to override or enable more limits. Note that
    limit names are invariant to use of dashes versus underscores.

    For more information on limits:

    * ``renderer.device.adapter.limits`` for the (max) limits available on the current system.
    * ``renderer.device.limits`` for the currently set limits.
    * https://gpuweb.github.io/gpuweb/#limits for the official webgpu limits.
    """
    if Shared._instance is not None:
        raise RuntimeError(
            "The set_wgpu_limits() function must be called before creating the first renderer."
        )
    Shared._limits.update(limits)


def get_shared():
    """Get the globally shared instance.

    Creates it if it does not yet exist. This should not be called at the import
    time of any module. Use this to get the global device:
    ``get_shared().device``.
    """
    if Shared._instance is None:
        Shared()

    return Shared._instance


class PyGfxAdapterInfoDiagnostics(wgpu.DiagnosticsBase):
    def get_dict(self):
        shared = get_shared()
        adapter = shared.adapter
        return adapter.info


class PyGfxFeaturesDiagnostics(wgpu.DiagnosticsBase):
    def get_dict(self):
        shared = get_shared()
        adapter = shared.adapter
        device = shared.device

        feature_names = list(wgpu.FeatureName)
        feature_names += sorted(adapter.features.difference(wgpu.FeatureName))
        result = {}
        for key in feature_names:
            result[key] = {
                "adapter": key in adapter.features,
                "device": key in device.features,
            }
        return result


class PyGfxLimitsDiagnostics(wgpu.DiagnosticsBase):
    def get_dict(self):
        shared = get_shared()
        adapter = shared.adapter
        device = shared.device

        result = {}
        for key in adapter.limits.keys():
            result[key] = {
                "adapter": adapter.limits[key],
                "device": device.limits.get(key, False),
            }
        return result


class PygfxCacheDiagnostics(wgpu.DiagnosticsBase):
    def get_dict(self):
        result = {}
        for cache_name, stats in gpu_caches.get_stats().items():
            count, hits, misses = stats
            result[cache_name] = {"count": count, "hits": hits, "misses": misses}
        return result


class PygfxResourceDiagnostics(wgpu.DiagnosticsBase):
    def get_dict(self):
        return Resource._resource_counts


def print_wgpu_report():
    """Print a report on the internal status of WGPU. Can be useful
    in debugging, and for providing details when making a bug report.
    """
    wgpu.diagnostics.print_report()
